import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class ParkingPosition(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.latitude = AddressableAttribute(localAddress='latitude', parent=self, value=None, valueType=float)
        self.longitude = AddressableAttribute(localAddress='longitude', parent=self, value=None, valueType=float)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update ParkingPosition from dict')

        # rename dict key to match new structure
        if 'data' in fromDict:
            fromDict['value'] = fromDict['data']
            del fromDict['data']

        if 'value' in fromDict:
            self.latitude.fromDict(fromDict['value'], 'lat')
            self.longitude.fromDict(fromDict['value'], 'lon')
        else:
            self.latitude.enabled = False
            self.longitude.enabled = False
            self.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['lat', 'lon']))

    def __str__(self):
        string = super().__str__()
        if self.latitude.enabled:
            string += f'\n\tLatitude: {self.latitude.value}'
        if self.longitude.enabled:
            string += f'\n\tLongitude: {self.longitude.value}'
        return string



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class BatteryStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.currentSOC_pct = AddressableAttribute(
            localAddress='currentSOC_pct', parent=self, value=None, valueType=int)
        self.cruisingRangeElectric_km = AddressableAttribute(
            localAddress='cruisingRangeElectric_km', value=None, parent=self, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update battery status from dict')

        if 'value' in fromDict:
            self.currentSOC_pct.fromDict(fromDict['value'], 'currentSOC_pct')

            if 'cruisingRangeElectric_km' in fromDict['value']:
                cruisingRangeElectric_km = int(fromDict['value']['cruisingRangeElectric_km'])
                if self.fixAPI and cruisingRangeElectric_km == 0x3FFF:
                    cruisingRangeElectric_km = None
                    LOG.info('%s: Attribute cruisingRangeElectric_km was error value 0x3FFF. Setting error state instead'
                             ' of 16383 km.', self.getGlobalAddress())
                self.cruisingRangeElectric_km.setValueWithCarTime(
                    cruisingRangeElectric_km, lastUpdateFromCar=None, fromServer=True)
            else:
                self.cruisingRangeElectric_km.enabled = False
        else:
            self.currentSOC_pct.enabled = False
            self.cruisingRangeElectric_km.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['currentSOC_pct', 'cruisingRangeElectric_km']))

    def __str__(self):
        string = super().__str__()
        if self.currentSOC_pct.enabled:
            string += f'\n\tCurrent SoC: {self.currentSOC_pct.value}%'
        if self.cruisingRangeElectric_km.enabled:
            if self.cruisingRangeElectric_km.value is not None:
                string += f'\n\tRange: {self.cruisingRangeElectric_km.value}km'
            else:
                string += '\n\tRange: currently unknown'
        return string



from enum import Enum
import re
import logging

import json
import requests

from weconnect.elements.error import Error
from weconnect.errors import SetterError
from weconnect.domain import Domain

from weconnect.addressable import AddressableLeaf, ChangeableAttribute, AddressableAttribute, AliasChangeableAttribute
from weconnect.elements.generic_settings import GenericSettings
from weconnect.util import celsiusToKelvin, farenheitToKelvin

LOG = logging.getLogger("weconnect")


class ClimatizationSettings(GenericSettings):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.targetTemperature_K = ChangeableAttribute(
            localAddress='targetTemperature_K', parent=self, value=None, valueType=float)
        self.targetTemperature_C = AliasChangeableAttribute(localAddress='targetTemperature_C', parent=self, value=None,
                                                            targetAttribute=self.targetTemperature_K, conversion=celsiusToKelvin, valueType=float)
        self.targetTemperature_F = AliasChangeableAttribute(localAddress='targetTemperature_F', parent=self, value=None,
                                                            targetAttribute=self.targetTemperature_K, conversion=farenheitToKelvin, valueType=float)
        self.unitInCar = AddressableAttribute(
            localAddress='unitInCar', parent=self, value=None, valueType=ClimatizationSettings.UnitInCar)
        self.climatisationWithoutExternalPower = ChangeableAttribute(
            localAddress='climatisationWithoutExternalPower', parent=self, value=None, valueType=bool)
        self.climatizationAtUnlock = ChangeableAttribute(
            localAddress='climatizationAtUnlock', parent=self, value=None, valueType=bool)
        self.windowHeatingEnabled = ChangeableAttribute(
            localAddress='windowHeatingEnabled', parent=self, value=None, valueType=bool)
        self.zoneFrontLeftEnabled = ChangeableAttribute(
            localAddress='zoneFrontLeftEnabled', parent=self, value=None, valueType=bool)
        self.zoneFrontRightEnabled = ChangeableAttribute(
            localAddress='zoneFrontRightEnabled', parent=self, value=None, valueType=bool)
        self.zoneRearLeftEnabled = ChangeableAttribute(
            localAddress='zoneRearLeftEnabled', parent=self, value=None, valueType=bool)
        self.zoneRearRightEnabled = ChangeableAttribute(
            localAddress='zoneRearRightEnabled', parent=self, value=None, valueType=bool)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

        self.targetTemperature_K.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.climatisationWithoutExternalPower.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.climatizationAtUnlock.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.windowHeatingEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneFrontLeftEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneFrontRightEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneRearLeftEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.zoneRearRightEnabled.addObserver(
            self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)

    def update(self, fromDict, ignoreAttributes=None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Climatization settings from dict')

        if 'value' in fromDict:
            self.targetTemperature_K.fromDict(fromDict['value'], 'targetTemperature_K')
            self.targetTemperature_C.fromDict(fromDict['value'], 'targetTemperature_C')
            self.targetTemperature_F.fromDict(fromDict['value'], 'targetTemperature_F')
            self.unitInCar.fromDict(fromDict['value'], 'unitInCar')
            self.climatisationWithoutExternalPower.fromDict(fromDict['value'], 'climatisationWithoutExternalPower')
            self.climatizationAtUnlock.fromDict(fromDict['value'], 'climatizationAtUnlock')
            self.windowHeatingEnabled.fromDict(fromDict['value'], 'windowHeatingEnabled')
            self.zoneFrontLeftEnabled.fromDict(fromDict['value'], 'zoneFrontLeftEnabled')
            self.zoneFrontRightEnabled.fromDict(fromDict['value'], 'zoneFrontRightEnabled')
            self.zoneRearLeftEnabled.fromDict(fromDict['value'], 'zoneRearLeftEnabled')
            self.zoneRearRightEnabled.fromDict(fromDict['value'], 'zoneRearRightEnabled')
        else:
            self.targetTemperature_K.enabled = False
            self.targetTemperature_C.enabled = False
            self.targetTemperature_F.enabled = False
            self.unitInCar.enabled = False
            self.climatisationWithoutExternalPower.enabled = False
            self.climatizationAtUnlock.enabled = False
            self.windowHeatingEnabled.enabled = False
            self.zoneFrontLeftEnabled.enabled = False
            self.zoneFrontRightEnabled.enabled = False
            self.zoneRearLeftEnabled.enabled = False
            self.zoneRearRightEnabled.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + [
            'targetTemperature_K',
            'targetTemperature_C',
            'targetTemperature_F',
            'unitInCar',
            'climatisationWithoutExternalPower',
            'climatizationAtUnlock',
            'windowHeatingEnabled',
            'zoneFrontLeftEnabled',
            'zoneFrontRightEnabled',
            'zoneRearLeftEnabled',
            'zoneRearRightEnabled']))

    def __str__(self):  # noqa: C901
        string = super().__str__()
        if self.targetTemperature_C.enabled:
            string += f'\n\tTarget Temperature in °C: {self.targetTemperature_C.value} °C '
        if self.targetTemperature_F.enabled:
            string += f'\n\tTarget Temperature in °F: {self.targetTemperature_F.value} °F '
        if self.targetTemperature_K.enabled:
            string += f'\n\tTarget Temperature in °K: {self.targetTemperature_K.value} °K '
        if self.unitInCar.enabled:
            string += f'\n\tTemperature unit in car: {self.unitInCar.value.value}'
        if self.climatisationWithoutExternalPower.enabled:
            string += f'\n\tClimatization without external Power: {self.climatisationWithoutExternalPower.value}'
        if self.climatizationAtUnlock.enabled:
            string += f'\n\tStart climatization after unlock: {self.climatizationAtUnlock.value}'
        if self.windowHeatingEnabled.enabled:
            string += f'\n\tWindow heating: {self.windowHeatingEnabled.value}'
        if self.zoneFrontLeftEnabled.enabled:
            string += f'\n\tHeating Front Left Seat: {self.zoneFrontLeftEnabled.value}'
        if self.zoneFrontRightEnabled.enabled:
            string += f'\n\tHeating Front Right Seat: {self.zoneFrontRightEnabled.value}'
        if self.zoneRearLeftEnabled.enabled:
            string += f'\n\tHeating Rear Left Seat: {self.zoneRearLeftEnabled.value}'
        if self.zoneRearRightEnabled.enabled:
            string += f'\n\tHeating Rear Right Seat: {self.zoneRearRightEnabled.value}'
        return string

    def valueChanged(self, element, flags):  # noqa: C901
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED \
                and not flags & AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER:
            url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/climatisation/settings'
            settingsDict = dict()
            regex = re.compile('zone(Front|Rear)(Right|Left)Enabled')
            if self.targetTemperature_C.enabled:
                settingsDict['targetTemperature'] = self.targetTemperature_C.value
                settingsDict['targetTemperatureUnit'] = ClimatizationSettings.UnitInCar.CELSIUS.value
            elif self.targetTemperature_F.enabled:
                settingsDict['targetTemperature'] = self.targetTemperature_F.value
                settingsDict['targetTemperatureUnit'] = ClimatizationSettings.UnitInCar.FARENHEIT.value
            else:
                settingsDict['targetTemperature'] = 20.0
                settingsDict['targetTemperatureUnit'] = ClimatizationSettings.UnitInCar.CELSIUS.value

            if self.climatisationWithoutExternalPower.enabled:
                settingsDict['climatisationWithoutExternalPower'] = self.climatisationWithoutExternalPower.value
            if self.climatizationAtUnlock.enabled:
                settingsDict['climatizationAtUnlock'] = self.climatizationAtUnlock.value
            if self.windowHeatingEnabled.enabled:
                settingsDict['windowHeatingEnabled'] = self.windowHeatingEnabled.value
            for child in self.getLeafChildren():
                if re.match(regex, child.getLocalAddress()):
                    settingsDict[child.getLocalAddress()] = child.value
            data = json.dumps(settingsDict)
            putResponse = self.vehicle.weConnect.session.put(url, data=data, allow_redirects=True)
            if putResponse.status_code != requests.codes['ok']:
                errorDict = putResponse.json()
                if errorDict is not None and 'error' in errorDict:
                    error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                    if error is not None:
                        message = ''
                        if error.message.enabled and error.message.value is not None:
                            message += error.message.value
                        if error.info.enabled and error.info.value is not None:
                            message += ' - ' + error.info.value
                        if error.retry.enabled and error.retry.value is not None:
                            if error.retry.value:
                                message += ' - Please retry in a moment'
                            else:
                                message += ' - No retry possible'
                        raise SetterError(f'Could not set value ({message})')
                    else:
                        raise SetterError(f'Could not set value ({putResponse.status_code})')
                raise SetterError(f'Could not not set value ({putResponse.status_code})')
            responseDict = putResponse.json()
            if 'data' in responseDict and 'requestID' in responseDict['data']:
                if self.vehicle.requestTracker is not None:
                    self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.ALL, 20, 120)

    class UnitInCar(Enum,):
        CELSIUS = 'celsius'
        FARENHEIT = 'farenheit'
        UNKNOWN = 'unknown unit'



# import os
# import json

# import pytest

# from weconnect import elements

# SCRIPTLOC = os.path.dirname(__file__)


def test_dummy():
    assert True

# @pytest.mark.parametrize('className, testcase', [(elements.ClimatizationTimer, 'complete')])
# def test_Elements(request, className, testcase):
#     with open(f'{request.config.rootdir}/tests/ressources/elements/{className.__name__}/{testcase}.json') as file:
#         testcaseDict = json.load(file)
#         element = className(parent=None, statusId='test', fromDict=testcaseDict)

#         element2 = className(parent=None, statusId='test', fromDict=None)
#         element2.update(testcaseDict)
#         print(repr(str(element2)))

#         assert element is not None

#         with open(f'{request.config.rootdir}/tests/ressources/elements/{className.__name__}/{testcase}.resultstr') \
#                 as resultStringFile:
#             expectedResult = resultStringFile.read()

#             assert expectedResult == str(element2)



import logging
import json
import requests

from weconnect.addressable import AddressableObject, ChangeableAttribute
from weconnect.elements.control_operation import ControlOperation, AccessControlOperation
from weconnect.elements.charging_settings import ChargingSettings
from weconnect.elements.climatization_settings import ClimatizationSettings
from weconnect.elements.error import Error
from weconnect.elements.window_heating_status import WindowHeatingStatus
from weconnect.elements.access_status import AccessStatus
from weconnect.errors import ControlError, SetterError
from weconnect.util import celsiusToKelvin, farenheitToKelvin
from weconnect.domain import Domain

LOG = logging.getLogger("weconnect")


class Controls(AddressableObject):
    def __init__(
        self,
        localAddress,
        vehicle,
        parent,
    ):
        self.vehicle = vehicle
        super().__init__(localAddress=localAddress, parent=parent)
        self.update()
        self.climatizationControl = None
        self.chargingControl = None
        self.windowHeatingControl = None
        self.accessControl = None
        self.wakeupControl = ChangeableAttribute(localAddress='wakeup', parent=self, value=ControlOperation.NONE, valueType=ControlOperation,
                                                 valueSetter=self.__setWakeupControlChange)

    def update(self):  # noqa: C901
        for domain in self.vehicle.domains.values():
            for status in domain.values():
                if isinstance(status, ClimatizationSettings) and not status.error.enabled:
                    if self.climatizationControl is None:
                        self.climatizationControl = ChangeableAttribute(
                            localAddress='climatisation', parent=self, value=ControlOperation.NONE, valueType=(ControlOperation, float),
                            valueSetter=self.__setClimatizationControlChange)
                elif isinstance(status, ChargingSettings) and not status.error.enabled:
                    if self.chargingControl is None:
                        self.chargingControl = ChangeableAttribute(
                            localAddress='charging', parent=self, value=ControlOperation.NONE, valueType=ControlOperation,
                            valueSetter=self.__setChargingControlChange)
                elif isinstance(status, WindowHeatingStatus) and not status.error.enabled:
                    if self.windowHeatingControl is None:
                        self.windowHeatingControl = ChangeableAttribute(
                            localAddress='windowheating', parent=self, value=ControlOperation.NONE, valueType=ControlOperation,
                            valueSetter=self.__setWindowHeatingControlChange)
                elif isinstance(status, AccessStatus) and not status.error.enabled and self.vehicle.weConnect.spin is not None \
                        and type(self.vehicle.weConnect.spin) != bool:
                    if self.accessControl is None:
                        self.accessControl = ChangeableAttribute(
                            localAddress='access', parent=self, value=AccessControlOperation.NONE, valueType=AccessControlOperation,
                            valueSetter=self.__setAccessControlChange)

    def __setClimatizationControlChange(self, value):  # noqa: C901
        if isinstance(value, ControlOperation):
            if value not in [ControlOperation.START, ControlOperation.STOP]:
                raise ControlError('Could not control climatisation, control operation %s cannot be executed', value)
            control = value
            temperature = None
        elif isinstance(value, (int, float)):
            control = ControlOperation.START
            temperature = float(value)
        else:
            raise ControlError('Could not control climatisation, control argument %s cannot be understood', value)

        url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/climatisation/{control.value}'

        settingsDict = dict()
        if control == ControlOperation.START:
            if 'climatisation' not in self.vehicle.domains and 'climatisationSettings' not in self.vehicle.domains['climatisation']:
                raise ControlError('Could not control climatisation, there are no climatisationSettings for the vehicle available.')
            climatizationSettings = self.vehicle.domains['climatisation']['climatisationSettings']
            for child in climatizationSettings.getLeafChildren():
                if isinstance(child, ChangeableAttribute):
                    settingsDict[child.getLocalAddress()] = child.value
            if temperature is not None:
                if 'targetTemperature_C' in settingsDict:
                    settingsDict['targetTemperature_C'] = temperature
                settingsDict['targetTemperature_K'] = celsiusToKelvin(temperature)
            elif 'targetTemperature_K' not in settingsDict:
                if 'targetTemperature_C' in settingsDict:
                    settingsDict['targetTemperature_K'] = celsiusToKelvin(settingsDict['targetTemperature_C'])
                elif 'targetTemperature_F' in settingsDict:
                    settingsDict['targetTemperature_K'] = farenheitToKelvin(settingsDict['targetTemperature_F'])
                else:
                    settingsDict['targetTemperature_K'] = celsiusToKelvin(20.5)

        data = json.dumps(settingsDict)
        controlResponse = self.vehicle.weConnect.session.post(url, data=data, allow_redirects=True)
        if controlResponse.status_code != requests.codes['ok']:
            errorDict = controlResponse.json()
            if errorDict is not None and 'error' in errorDict:
                error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                if error is not None:
                    message = ''
                    if error.message.enabled and error.message.value is not None:
                        message += error.message.value
                    if error.info.enabled and error.info.value is not None:
                        message += ' - ' + error.info.value
                    if error.retry.enabled and error.retry.value is not None:
                        if error.retry.value:
                            message += ' - Please retry in a moment'
                        else:
                            message += ' - No retry possible'
                    raise SetterError(f'Could not control climatisation ({message})')
                else:
                    raise SetterError(f'Could not control climatisation ({controlResponse.status_code})')
            raise SetterError(f'Could not control climatisation ({controlResponse.status_code})')
        responseDict = controlResponse.json()
        if 'data' in responseDict and 'requestID' in responseDict['data']:
            if self.vehicle.requestTracker is not None:
                self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.CLIMATISATION, 20, 120)

    def __setChargingControlChange(self, value):  # noqa: C901
        if value in [ControlOperation.START, ControlOperation.STOP]:
            url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/charging/{value.value}'

            controlResponse = self.vehicle.weConnect.session.post(url, data='{}', allow_redirects=True)
            if controlResponse.status_code != requests.codes['ok']:
                errorDict = controlResponse.json()
                if errorDict is not None and 'error' in errorDict:
                    error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                    if error is not None:
                        message = ''
                        if error.message.enabled and error.message.value is not None:
                            message += error.message.value
                        if error.info.enabled and error.info.value is not None:
                            message += ' - ' + error.info.value
                        if error.retry.enabled and error.retry.value is not None:
                            if error.retry.value:
                                message += ' - Please retry in a moment'
                            else:
                                message += ' - No retry possible'
                        raise SetterError(f'Could not control charging ({message})')
                    else:
                        raise SetterError(f'Could not control charging ({controlResponse.status_code})')
                raise SetterError(f'Could not control charging ({controlResponse.status_code})')
            responseDict = controlResponse.json()
            if 'data' in responseDict and 'requestID' in responseDict['data']:
                if self.vehicle.requestTracker is not None:
                    self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.CHARGING, 20, 120)

    def __setWindowHeatingControlChange(self, value):  # noqa: C901
        if value in [ControlOperation.START, ControlOperation.STOP]:
            url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/windowheating/{value.value}'

            controlResponse = self.vehicle.weConnect.session.post(url, data='{}', allow_redirects=True)
            if controlResponse.status_code != requests.codes['ok']:
                errorDict = controlResponse.json()
                if errorDict is not None and 'error' in errorDict:
                    error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                    if error is not None:
                        message = ''
                        if error.message.enabled and error.message.value is not None:
                            message += error.message.value
                        if error.info.enabled and error.info.value is not None:
                            message += ' - ' + error.info.value
                        if error.retry.enabled and error.retry.value is not None:
                            if error.retry.value:
                                message += ' - Please retry in a moment'
                            else:
                                message += ' - No retry possible'
                        raise SetterError(f'Could not control windowheating ({message})')
                    else:
                        raise SetterError(f'Could not control windowheating ({controlResponse.status_code})')
                raise SetterError(f'Could not control windowheating ({controlResponse.status_code})')
            responseDict = controlResponse.json()
            if 'data' in responseDict and 'requestID' in responseDict['data']:
                if self.vehicle.requestTracker is not None:
                    self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.CLIMATISATION, 20, 120)

    def __setWakeupControlChange(self, value):  # noqa: C901
        if value in [ControlOperation.START]:
            url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/vehiclewakeuptrigger'

            controlResponse = self.vehicle.weConnect.session.post(url, data='{}', allow_redirects=True)

            if controlResponse.status_code not in (requests.codes['ok'], requests.codes['no_content']):
                errorDict = controlResponse.json()
                if errorDict is not None and 'error' in errorDict:
                    error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                    if error is not None:
                        message = ''
                        if error.message.enabled and error.message.value is not None:
                            message += error.message.value
                        if error.info.enabled and error.info.value is not None:
                            message += ' - ' + error.info.value
                        if error.retry.enabled and error.retry.value is not None:
                            if error.retry.value:
                                message += ' - Please retry in a moment'
                            else:
                                message += ' - No retry possible'
                        raise SetterError(f'Could not control wakeup ({message})')
                    else:
                        raise SetterError(f'Could not control wakeup ({controlResponse.status_code})')
                raise SetterError(f'Could not control wakeup ({controlResponse.status_code})')

    def __setAccessControlChange(self, value):  # noqa: C901
        if isinstance(value, AccessControlOperation):
            if value not in [AccessControlOperation.LOCK, AccessControlOperation.UNLOCK]:
                raise ControlError('Could not control access, control operation %s cannot be executed', value)
            control = value
        else:
            raise ControlError('Could not control access, control argument %s cannot be understood', value)
        if self.vehicle.weConnect.spin is None or type(self.vehicle.weConnect.spin) != str:
            raise ControlError('Could not control access, control operation needs an S-PIN')
        spin = self.vehicle.weConnect.spin

        url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/access/{control.value}'

        data = {}
        data['spin'] = spin
        controlResponse = self.vehicle.weConnect.session.post(url, json=data, allow_redirects=True)
        if controlResponse.status_code != requests.codes['ok']:
            errorDict = controlResponse.json()
            if errorDict is not None and 'error' in errorDict:
                error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                if error is not None:
                    message = ''
                    if error.message.enabled and error.message.value is not None:
                        message += error.message.value
                    if error.info.enabled and error.info.value is not None:
                        message += ' - ' + error.info.value
                    if error.retry.enabled and error.retry.value is not None:
                        if error.retry.value:
                            message += ' - Please retry in a moment'
                        else:
                            message += ' - No retry possible'
                    raise SetterError(f'Could not control access ({message})')
                else:
                    raise SetterError(f'Could not control access ({controlResponse.status_code})')
            raise SetterError(f'Could not control access ({controlResponse.status_code})')



from datetime import datetime
import logging

from weconnect.addressable import AddressableAttribute, AddressableDict
from weconnect.elements.generic_status import GenericStatus
from weconnect.elements.timer import Timer

LOG = logging.getLogger("weconnect")


class ClimatizationTimer(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.timers = AddressableDict(localAddress='timers', parent=self)
        self.timeInCar = AddressableAttribute(localAddress='timeInCar', parent=self, value=None, valueType=datetime)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update climatization timer from dict')

        if 'value' in fromDict:
            if 'timers' in fromDict['value'] and fromDict['value']['timers'] is not None:
                for climatizationTimerDict in fromDict['value']['timers']:
                    if 'id' in climatizationTimerDict:
                        if climatizationTimerDict['id'] in self.timers:
                            self.timers[climatizationTimerDict['id']].update(fromDict=climatizationTimerDict)
                        else:
                            self.timers[climatizationTimerDict['id']] = Timer(
                                fromDict=climatizationTimerDict, parent=self.timers)
                for timerId in [timerId for timerId in self.timers.keys()
                                if timerId not in [timer['id']
                                for timer in fromDict['value']['timers'] if 'id' in timer]]:
                    del self.timers[timerId]
            else:
                self.timers.clear()
                self.timers.enabled = False

            self.timeInCar.fromDict(fromDict['value'], 'timeInCar')
        else:
            self.timers.clear()
            self.timers.enabled = False
            self.timeInCar.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['timers', 'timeInCar']))

    def __str__(self):
        string = super().__str__()
        if self.timeInCar.enabled:
            string += f'\n\tTime in Car: {self.timeInCar.value.isoformat()}'  # pylint: disable=no-member
            string += f' (captured at {self.carCapturedTimestamp.value.isoformat()})'  # pylint: disable=no-member
        string += f'\n\tTimers: {len(self.timers)} items'
        for timer in self.timers.values():
            string += f'\n\t\t{timer}'
        return string



import logging

from weconnect.addressable import AddressableDict
from weconnect.elements.generic_status import GenericStatus
from weconnect.elements.generic_capability import GenericCapability

LOG = logging.getLogger("weconnect")


class CapabilityStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.capabilities = AddressableDict(localAddress='capabilities', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update capability status from dict')

        if 'value' in fromDict:
            for capDict in fromDict['value']:
                if 'id' in capDict:
                    if capDict['id'] in self.capabilities:
                        self.capabilities[capDict['id']].update(fromDict=capDict)
                    else:
                        self.capabilities[capDict['id']] = GenericCapability(
                            capabilityId=capDict['id'], fromDict=capDict, parent=self.capabilities)
            for capabilityId in [capabilityId for capabilityId in self.capabilities.keys()
                                 if capabilityId not in [capability['id']
                                 for capability in fromDict['value'] if 'id' in capability]]:
                del self.capabilities[capabilityId]
        else:
            self.capabilities.clear()
            self.capabilities.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes))

    def __str__(self):
        string = super().__str__()
        string += f'\n\tCapabilities: {len(self.capabilities)} items'
        for capability in self.capabilities.values():
            string += f'\n\t\t{capability}'
        return string



from enum import Enum
from datetime import datetime
import logging

from weconnect.addressable import AddressableAttribute, AddressableDict, AddressableObject
from weconnect.elements.generic_status import GenericStatus

from weconnect.util import robustTimeParse, toBool

LOG = logging.getLogger("weconnect")


class DepartureTimersStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.heaterSource = AddressableAttribute(localAddress='heaterSource', parent=self, value=None, valueType=DepartureTimersStatus.HeaterSource)
        self.minSOC_pct = AddressableAttribute(localAddress='minSOC_pct', parent=self, value=None, valueType=int)
        self.timers = AddressableDict(localAddress='timers', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update climatization timer from dict')

        if 'value' in fromDict:
            self.heaterSource.fromDict(fromDict['value'], 'heaterSource')
            self.minSOC_pct.fromDict(fromDict['value'], 'minSOC_pct')

            if 'timers' in fromDict['value'] and fromDict['value']['timers'] is not None:
                for climatizationTimerDict in fromDict['value']['timers']:
                    if 'id' in climatizationTimerDict:
                        if climatizationTimerDict['id'] in self.timers:
                            self.timers[climatizationTimerDict['id']].update(fromDict=climatizationTimerDict)
                        else:
                            self.timers[climatizationTimerDict['id']] = DepartureTimersStatus.Timer(
                                fromDict=climatizationTimerDict, parent=self.timers)
                for timerId in [timerId for timerId in self.timers.keys()
                                if timerId not in [timer['id']
                                for timer in fromDict['value']['timers'] if 'id' in timer]]:
                    del self.timers[timerId]
            else:
                self.timers.clear()
                self.timers.enabled = False

        else:
            self.heaterSource.enabled = False
            self.minSOC_pct.enabled = False
            self.timers.clear()
            self.timers.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['timers', 'heaterSource', 'minSOC_pct']))

    def __str__(self):
        string = super().__str__()
        if self.heaterSource.enabled:
            string += f'\n\tHeater Source: {self.heaterSource.value.value}'
        if self.minSOC_pct.enabled:
            string += f'\n\tMin SoC: {self.minSOC_pct.value}%'
        string += f'\n\tTimers: {len(self.timers)} items\n'
        for timer in self.timers.values():
            string += ''.join(['\t\t' + line for line in str(timer).splitlines(True)]) + '\n'
        return string

    class HeaterSource(Enum,):
        ELECTRIC = 'electric'
        UNKNOWN = 'unknown'

    class Timer(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.timerEnabled = AddressableAttribute(localAddress='enabled', parent=self, value=None, valueType=bool)
            self.climatisation = AddressableAttribute(localAddress='climatisation', parent=self, value=None, valueType=bool)
            self.charging = AddressableAttribute(localAddress='charging', parent=self, value=None, valueType=bool)
            self.recurringTimer = None
            self.singleTimer = None
            self.preferredChargingTimes = AddressableDict(localAddress='preferredChargingTimes', parent=self)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):  # noqa: C901
            LOG.debug('Update timer from dict')

            if 'id' in fromDict:
                self.id = fromDict['id']
                self.localAddress = str(self.id)
            else:
                LOG.error('Timer is missing id attribute')

            if 'enabled' in fromDict:
                self.timerEnabled.setValueWithCarTime(toBool(fromDict['enabled']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.timerEnabled.enabled = False

            if 'climatisation' in fromDict:
                self.climatisation.setValueWithCarTime(toBool(fromDict['climatisation']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.climatisation.enabled = False

            if 'charging' in fromDict:
                self.charging.setValueWithCarTime(toBool(fromDict['charging']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.charging.enabled = False

            if 'recurringTimer' in fromDict:
                if self.recurringTimer is None:
                    self.recurringTimer = DepartureTimersStatus.Timer.RecurringTimer(
                        localAddress='recurringTimer', parent=self, fromDict=fromDict['recurringTimer'])
                else:
                    self.recurringTimer.update(fromDict=fromDict['recurringTimer'])
            elif self.recurringTimer is not None:
                self.recurringTimer.enabled = False
                self.recurringTimer = None

            if 'singleTimer' in fromDict:
                if self.singleTimer is None:
                    self.singleTimer = DepartureTimersStatus.Timer.SingleTimer(
                        localAddress='singleTimer', parent=self, fromDict=fromDict['singleTimer'])
                else:
                    self.singleTimer.update(fromDict=fromDict['singleTimer'])
            elif self.singleTimer is not None:
                self.singleTimer.enabled = False
                self.singleTimer = None

            if 'preferredChargingTimes' in fromDict and fromDict['preferredChargingTimes'] is not None:
                for preferredChargingTimeDict in fromDict['preferredChargingTimes']:
                    if 'id' in preferredChargingTimeDict:
                        if preferredChargingTimeDict['id'] in self.preferredChargingTimes:
                            self.preferredChargingTimes[preferredChargingTimeDict['id']].update(fromDict=preferredChargingTimeDict)
                        else:
                            self.preferredChargingTimes[preferredChargingTimeDict['id']] = DepartureTimersStatus.Timer.PreferredChargingTimes(
                                localAddress=preferredChargingTimeDict['id'], fromDict=preferredChargingTimeDict, parent=self.preferredChargingTimes)
                for timerId in [timerId for timerId in self.preferredChargingTimes.keys()
                                if timerId not in [timer['id']
                                for timer in fromDict['preferredChargingTimes'] if 'id' in timer]]:
                    del self.PreferredChargingTimes[timerId]
            else:
                self.preferredChargingTimes.clear()
                self.preferredChargingTimes.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['id', 'enabled', 'climatisation', 'charging', 'recurringTimer', 'singleTimer', 'preferredChargingTimes']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            string = f'{self.id}: Enabled: {self.timerEnabled.value}'
            if self.climatisation.enabled:
                string += f', Climatisation: {self.climatisation.value}'
            if self.charging.enabled:
                string += f', Charging: {self.charging.value}'
            if self.recurringTimer is not None and self.recurringTimer.enabled:
                string += f' at {self.recurringTimer} '
            if self.singleTimer is not None and self.singleTimer.enabled:
                string += f' at {self.singleTimer} '
            if self.preferredChargingTimes.enabled and len(self.preferredChargingTimes) > 0:
                string += f'\n\tPreferred charging times: {len(self.preferredChargingTimes)} items'
                for preferredChargingTime in self.preferredChargingTimes.values():
                    string += f'\n\t\t{preferredChargingTime}'
            return string

        class RecurringTimer(AddressableObject):
            def __init__(
                self,
                localAddress,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress=localAddress, parent=parent)
                self.departureTimeLocal = AddressableAttribute(
                    localAddress='departureTimeLocal', parent=self, value=None, valueType=datetime)
                self.repetitionDays = AddressableDict(localAddress='repetitionDays', parent=self)
                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update recurring timer from dict')

                if 'departureTimeLocal' in fromDict:
                    self.departureTimeLocal.setValueWithCarTime(datetime.strptime(f'{fromDict["departureTimeLocal"]}+00:00', '%H:%M%z'),
                                                                lastUpdateFromCar=None, fromServer=True)
                else:
                    self.departureTimeLocal.enabled = False

                if 'repetitionDays' in fromDict and fromDict['repetitionDays'] is not None:
                    for day in fromDict['repetitionDays']:
                        if day in self.repetitionDays:
                            self.repetitionDays[day].setValueWithCarTime(True, lastUpdateFromCar=None, fromServer=True)
                        else:
                            self.repetitionDays[day] = AddressableAttribute(
                                localAddress=day, parent=self.repetitionDays, value=True, valueType=bool)
                    for day in [day for day in self.repetitionDays if day not in fromDict['repetitionDays']]:
                        del self.repetitionDays[day]
                else:
                    self.repetitionDays.clear()
                    self.repetitionDays.enabled = False

                for key, value in {key: value for key, value in fromDict.items()
                                   if key not in ['departureTimeLocal', 'repetitionDays']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                string = f'{self.departureTimeLocal.value.strftime("%H:%M")} on '  # pylint: disable=no-member
                for day, value in self.repetitionDays.items():
                    if value:
                        string += day + ' '
                return string

        class PreferredChargingTimes(AddressableObject):
            def __init__(
                self,
                localAddress,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress=localAddress, parent=parent)
                self.timerEnabled = AddressableAttribute(localAddress='enabled', parent=self, value=None, valueType=bool)
                self.startTimeLocal = AddressableAttribute(
                    localAddress='startTimeLocal', parent=self, value=None, valueType=datetime)
                self.endTimeLocal = AddressableAttribute(
                    localAddress='endTimeLocal', parent=self, value=None, valueType=datetime)
                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update preferred charging times from dict')

                if 'id' in fromDict:
                    self.id = fromDict['id']
                    self.localAddress = str(self.id)
                else:
                    LOG.error('Charging time is missing id attribute')

                if 'enabled' in fromDict:
                    self.timerEnabled.setValueWithCarTime(toBool(fromDict['enabled']), lastUpdateFromCar=None, fromServer=True)
                else:
                    self.timerEnabled.enabled = False

                if 'startTimeLocal' in fromDict:
                    self.startTimeLocal.setValueWithCarTime(datetime.strptime(f'{fromDict["startTimeLocal"]}+00:00', '%H:%M%z'),
                                                            lastUpdateFromCar=None, fromServer=True)
                else:
                    self.startTimeLocal.enabled = False

                if 'endTimeLocal' in fromDict:
                    self.endTimeLocal.setValueWithCarTime(datetime.strptime(f'{fromDict["endTimeLocal"]}+00:00', '%H:%M%z'),
                                                          lastUpdateFromCar=None, fromServer=True)
                else:
                    self.endTimeLocal.enabled = False

                for key, value in {key: value for key, value in fromDict.items()
                                   if key not in ['id', 'enabled', 'startTimeLocal', 'endTimeLocal']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                string = f'{self.id}: Enabled: {self.timerEnabled.value}'
                if self.startTimeLocal.enabled:
                    string += f', Start: {self.startTimeLocal.value.strftime("%H:%M")}'
                if self.endTimeLocal.enabled:
                    string += f', End: {self.endTimeLocal.value.strftime("%H:%M")}'
                return string

        class SingleTimer(AddressableObject):
            def __init__(
                self,
                localAddress,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress=localAddress, parent=parent)
                self.departureDateTimeLocal = AddressableAttribute(
                    localAddress='departureDateTimeLocal', parent=self, value=None, valueType=datetime)
                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update recurring timer from dict')

                if 'departureDateTimeLocal' in fromDict:
                    self.departureDateTimeLocal.setValueWithCarTime(robustTimeParse(fromDict["departureDateTimeLocal"]), lastUpdateFromCar=None,
                                                                    fromServer=True)
                else:
                    self.departureDateTimeLocal.enabled = False

                for key, value in {key: value for key, value in fromDict.items()
                                   if key not in ['departureDateTimeLocal']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                returnString = ""
                if self.departureDateTimeLocal.enabled:
                    returnString += self.departureDateTimeLocal.value.isoformat()  # pylint: disable=no-member
                return returnString



import logging
from enum import Enum

from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableList

LOG = logging.getLogger("weconnect")


class ChargingStation(AddressableObject):  # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        weConnect,
        stationId,
        parent,
        fromDict,
        fixAPI=True,
    ):
        self.weConnect = weConnect
        super().__init__(localAddress=stationId, parent=parent)
        self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=str)
        self.name = AddressableAttribute(localAddress='name', parent=self, value=None, valueType=str)
        self.latitude = AddressableAttribute(localAddress='latitude', parent=self, value=None, valueType=float)
        self.longitude = AddressableAttribute(localAddress='longitude', parent=self, value=None, valueType=float)
        self.distance = AddressableAttribute(localAddress='distance', parent=self, value=None, valueType=float)
        self.address = None
        self.chargingPower = AddressableAttribute(localAddress='chargingPower', parent=self, value=None, valueType=float)
        self.chargingSpots = AddressableList(localAddress='chargingSpots', parent=self)
        self.authTypes = AddressableList(localAddress='authTypes', parent=self)
        self.filteredOut = AddressableAttribute(localAddress='filteredOut', parent=self, value=None, valueType=bool)
        self.isFavorite = AddressableAttribute(localAddress='isFavorite', parent=self, value=None, valueType=bool)
        self.operator = None
        self.isWeChargePartner = AddressableAttribute(localAddress='isWeChargePartner', parent=self, value=None, valueType=bool)

        self.fixAPI = fixAPI

        self.update(fromDict)

    def update(  # noqa: C901  # pylint: disable=too-many-branches
        self,
        fromDict=None,
    ):
        if fromDict is not None:
            LOG.debug('Create / update charging station')
            self.id.fromDict(fromDict, 'id')
            self.name.fromDict(fromDict, 'name')
            self.latitude.fromDict(fromDict, 'latitude')
            self.longitude.fromDict(fromDict, 'longitude')
            self.distance.fromDict(fromDict, 'distance')

            if 'address' in fromDict:
                if self.address is None:
                    self.address = ChargingStation.Address(localAddress='address', parent=self, fromDict=fromDict['address'])
                else:
                    self.address.update(fromDict=fromDict['address'])
            elif self.address is not None:
                self.address.enabled = False
                self.address = None

            self.chargingPower.fromDict(fromDict, 'chargingPower')

            if 'chargingSpots' in fromDict and fromDict['chargingSpots'] is not None:
                if len(fromDict['chargingSpots']) == len(self.chargingSpots):
                    for i, spot in enumerate(fromDict['chargingSpots']):
                        self.chargingSpots[i].update(fromDict=spot)
                else:
                    self.chargingSpots.clear()
                    for spot in fromDict['chargingSpots']:
                        self.chargingSpots.append(ChargingStation.ChargingSpot(localAddress=str(
                            len(self.chargingSpots)), parent=self.chargingSpots, fromDict=spot))
            else:
                self.chargingSpots.enabled = False
                self.chargingSpots.clear()

            if 'authTypes' in fromDict and fromDict['authTypes'] is not None:
                if len(fromDict['authTypes']) == len(self.authTypes):
                    for i, authType in enumerate(fromDict['authTypes']):
                        try:
                            authTypeEnum = ChargingStation.AUTHTYPE(authType)
                        except ValueError:
                            authTypeEnum = ChargingStation.AUTHTYPE.UNKNOWN
                            LOG.warning('An unsupported type: %s was provided, please report this as a bug', authTypeEnum)
                        self.authTypes[i].setValueWithCarTime(authTypeEnum, lastUpdateFromCar=None, fromServer=True)
                else:
                    self.authTypes.clear()
                    for authType in fromDict['authTypes']:
                        try:
                            authTypeEnum = ChargingStation.AUTHTYPE(authType)
                        except ValueError:
                            authTypeEnum = ChargingStation.AUTHTYPE.UNKNOWN
                            LOG.warning('An unsupported type: %s was provided, please report this as a bug', authTypeEnum)
                        self.authTypes.append(AddressableAttribute(localAddress=len(self.authTypes),
                                              parent=self.authTypes, value=authTypeEnum, valueType=ChargingStation.AUTHTYPE))
            else:
                self.authTypes.enabled = False
                self.authTypes.clear()

            self.filteredOut.fromDict(fromDict, 'filteredOut')
            self.isFavorite.fromDict(fromDict, 'isFavorite')
            self.isWeChargePartner.fromDict(fromDict, 'isWeChargePartner')

            if 'cpoiOperatorInfo' in fromDict:
                if self.operator is None:
                    self.operator = ChargingStation.Operator(localAddress='operator', parent=self, fromDict=fromDict['cpoiOperatorInfo'])
                else:
                    self.operator.update(fromDict=fromDict['cpoiOperatorInfo'])
            elif self.operator is not None:
                self.operator.enabled = False
                self.operator = None

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['id',
                                              'name',
                                              'latitude',
                                              'longitude',
                                              'distance',
                                              'address',
                                              'chargingPower',
                                              'chargingSpots',
                                              'authTypes',
                                              'filteredOut',
                                              'isFavorite',
                                              'isWeChargePartner',
                                              'cpoiOperatorInfo']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

    class AUTHTYPE(Enum):
        RFID = 'RFID'
        APP = 'APP'
        QR = 'QR'
        NO_AUTH = 'NO_AUTH'
        UNKNOWN = 'UNKNOWN'

    class Address(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.city = AddressableAttribute(localAddress='city', parent=self, value=None, valueType=str)
            self.country = AddressableAttribute(localAddress='country', parent=self, value=None, valueType=str)
            self.postcode = AddressableAttribute(localAddress='postcode', parent=self, value=None, valueType=str)
            self.street = AddressableAttribute(localAddress='street', parent=self, value=None, valueType=str)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update address from dict')

            if 'city' in fromDict:
                self.city.setValueWithCarTime(fromDict['city'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.city.enabled = False

            if 'country' in fromDict:
                self.country.setValueWithCarTime(fromDict['country'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.country.enabled = False

            if 'postcode' in fromDict:
                self.postcode.setValueWithCarTime(fromDict['postcode'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.postcode.enabled = False

            if 'street' in fromDict:
                self.street.setValueWithCarTime(fromDict['street'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.street.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['city', 'country', 'postcode', 'street']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = ''
            if self.street.enabled:
                returnString += f'{self.street.value}, '
            if self.postcode.enabled:
                returnString += f'{self.postcode.value} '
            if self.city.enabled:
                returnString += f'{self.city.value}, '
            if self.country.enabled:
                returnString += f'{self.country.value}'
            return returnString

    class ChargingSpot(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.connectors = AddressableList(localAddress='connectors', parent=self)
            self.available = AddressableAttribute(localAddress='available', parent=self, value=None, valueType=ChargingStation.ChargingSpot.AVAILABILITY)
            self.chargingPower = AddressableAttribute(localAddress='chargingPower', parent=self, value=None, valueType=float)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update charging spot from dict')

            if 'connectors' in fromDict and fromDict['connectors'] is not None:
                if len(fromDict['connectors']) == len(self.connectors):
                    for i, connector in enumerate(fromDict['connectors']):
                        self.connectors[i].update(fromDict=connector)
                else:
                    self.connectors.clear()
                    for connector in fromDict['connectors']:
                        self.connectors.append(ChargingStation.ChargingSpot.Connector(
                            localAddress=str(len(self.connectors)), parent=self.connectors, fromDict=connector))
            else:
                self.connectors.enabled = False
                self.connectors.clear()

            if 'chargingPower' in fromDict:
                self.chargingPower.setValueWithCarTime(float(fromDict['chargingPower']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.chargingPower.enabled = False

            if 'available' in fromDict:
                try:
                    self.available.setValueWithCarTime(ChargingStation.ChargingSpot.AVAILABILITY(fromDict['available']),
                                                       lastUpdateFromCar=None, fromServer=True)
                except ValueError:
                    self.available.setValueWithCarTime(ChargingStation.ChargingSpot.AVAILABILITY.UNKNOWN,
                                                       lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('An unsupported type: %s was provided,'
                                ' please report this as a bug', fromDict['available'])
            else:
                self.available.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['connectors', 'chargingPower', 'available', 'plugTypes']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = ''
            if self.available.enabled:
                returnString += f'Availability: {self.available.value.value}\n'  # pylint: disable=no-member
            if self.chargingPower.enabled:
                returnString += f'Max. Charging Power: {self.chargingPower.value}kW\n'
            returnString += f'Connectors: {len(self.connectors)} items\n'
            for connector in self.connectors:
                returnString += ''.join(['\t' + line for line in str(connector).splitlines(True)]) + '\n'
            return returnString

        class PlugType(Enum):
            TYPE1 = 'Type1'
            TYPE2 = 'Type2'
            CHADEMO = 'CHAdeMO'
            CCS = 'CCS'
            SCHUKO = 'Schuko'
            CEE3 = 'CEE3'
            UNKNOWN = 'unknown'

        class AVAILABILITY(Enum):
            AVAILABLE = 'AVAILABLE'
            OCCUPIED = 'OCCUPIED'
            UNKNOWN = 'UNKNOWN'

        class Connector(AddressableObject):
            def __init__(
                self,
                localAddress,
                parent,
                fromDict=None,
            ):
                super().__init__(localAddress=localAddress, parent=parent)
                self.plugType = AddressableAttribute(localAddress='plugType', parent=self, value=None, valueType=ChargingStation.ChargingSpot.PlugType)
                self.chargingPower = AddressableAttribute(localAddress='chargingPower', parent=self, value=None, valueType=float)

                if fromDict is not None:
                    self.update(fromDict)

            def update(self, fromDict):
                LOG.debug('Update connector from dict')

                if 'plugType' in fromDict:
                    try:
                        self.plugType.setValueWithCarTime(ChargingStation.ChargingSpot.PlugType(fromDict['plugType']),
                                                          lastUpdateFromCar=None, fromServer=True)
                    except ValueError:
                        self.plugType.setValueWithCarTime(ChargingStation.ChargingSpot.PlugType.UNKNOWN,
                                                          lastUpdateFromCar=None, fromServer=True)
                        LOG.warning('An unsupported type: %s was provided,'
                                    ' please report this as a bug', fromDict['plugType'])
                else:
                    self.plugType.enabled = False

                if 'chargingPower' in fromDict:
                    self.chargingPower.setValueWithCarTime(float(fromDict['chargingPower']), lastUpdateFromCar=None, fromServer=True)
                else:
                    self.chargingPower.enabled = False

                for key, value in {key: value for key, value in fromDict.items()
                                   if key not in ['plugType', 'chargingPower']}.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

            def __str__(self):
                returnString = ''
                if self.plugType.enabled:
                    returnString += f'Plug Type: {self.plugType.value.value}\n'  # pylint: disable=no-member
                if self.chargingPower.enabled:
                    returnString += f'Max. Charging Power: {self.chargingPower.value}kW'
                return returnString

    class Operator(AddressableObject):
        def __init__(
            self,
            localAddress,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=localAddress, parent=parent)
            self.name = AddressableAttribute(localAddress='name', parent=self, value=None, valueType=str)
            self.id = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=str)
            self.phoneNumber = AddressableAttribute(localAddress='phoneNumber', parent=self, value=None, valueType=str)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update charging spot from dict')

            if 'name' in fromDict:
                self.name.setValueWithCarTime(fromDict['name'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.name.enabled = False

            if 'id' in fromDict:
                self.id.setValueWithCarTime(fromDict['id'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.id.enabled = False

            if 'phoneNumber' in fromDict:
                self.phoneNumber.setValueWithCarTime(fromDict['phoneNumber'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.phoneNumber.enabled = False

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['name', 'id', 'phoneNumber']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnString = ''
            if self.name.enabled:
                returnString += self.name.value
            if self.id.enabled:
                returnString += f' (Id: {self.id.value})'
            if self.phoneNumber.enabled and self.phoneNumber.value:
                returnString += f' Phone: {self.phoneNumber.value}'
            return returnString

    def __str__(self):  # noqa: C901
        returnString = ''
        if self.id.enabled:
            returnString += f'ID:                  {self.id.value}\n'
        if self.name.enabled:
            returnString += f'Name:                {self.name.value}\n'
        if self.operator is not None and self.operator.enabled:
            returnString += f'Operator:            {self.operator}\n'
        if self.latitude.enabled:
            returnString += f'Latitude:            {self.latitude.value}\n'
        if self.longitude.enabled:
            returnString += f'Longitude:           {self.longitude.value}\n'
        if self.distance.enabled:
            returnString += f'Distance:            {round(self.distance.value)}m\n'
        if self.address is not None and self.address.enabled:
            returnString += f'Address:             {self.address}\n'
        if self.chargingPower.enabled:
            returnString += f'Max. Charging Power: {self.chargingPower.value}kW\n'
        returnString += f'Charging Spots: {len(self.chargingSpots)} items\n'
        for spot in self.chargingSpots:
            returnString += ''.join(['\t' + line for line in str(spot).splitlines(True)]) + '\n'
        returnString += f'Authentification:    {", ".join([authtype.value.value for authtype in self.authTypes])}\n'
        returnString += 'Options:             '
        if self.filteredOut.enabled and self.filteredOut.value:
            returnString += 'filtered out; '
        if self.isFavorite.enabled and self.isFavorite.value:
            returnString += 'favourite; '
        if self.isWeChargePartner.enabled and self.isWeChargePartner.value:
            returnString += 'weCharge partner; '
        returnString += '\n'
        return returnString



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class LVBatteryStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.batterySupport = AddressableAttribute(localAddress='batterySupport', value=None, parent=self,
                                                   valueType=LVBatteryStatus.BatterySupport)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update lv battery status from dict')

        if 'value' in fromDict:
            self.batterySupport.fromDict(fromDict['value'], 'batterySupport')
        else:
            self.batterySupport.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['batterySupport']))

    def __str__(self):
        string = super().__str__()
        if self.batterySupport.enabled:
            string += f'\n\tBattery Support: {self.batterySupport.value.value}'  # pylint: disable=no-member
        return string

    class BatterySupport(Enum,):
        NOT_ALLOWED = 'notAllowed'
        PENDING_DISABLED = 'pendingDisable'
        DISABLED = 'disabled'
        PENDING_ENABLED = 'pendingEnable'
        ENABLED = 'enabled'
        INVALID = 'invalid'
        UNKNOWN = 'unknown battery support'



from enum import Enum
import logging

import json
import requests

from weconnect.addressable import AddressableLeaf, ChangeableAttribute, AliasChangeableAttribute
from weconnect.elements.generic_status import GenericStatus
from weconnect.elements.error import Error
from weconnect.errors import SetterError
from weconnect.domain import Domain

LOG = logging.getLogger("weconnect")


class GenericSettings(GenericStatus):
    def valueChanged(self, element, flags):  # noqa: C901
        del element
        if flags & AddressableLeaf.ObserverEvent.VALUE_CHANGED \
                and not flags & AddressableLeaf.ObserverEvent.UPDATED_FROM_SERVER:
            setting = self.id.partition('Settings')[0]
            url = f'https://mobileapi.apps.emea.vwapps.io/vehicles/{self.vehicle.vin.value}/{setting}/settings'
            settingsDict = dict()
            for child in self.getLeafChildren():
                if isinstance(child, ChangeableAttribute) and not isinstance(child, AliasChangeableAttribute):
                    if isinstance(child.value, Enum):  # pylint: disable=no-member # this is a fales positive
                        settingsDict[child.getLocalAddress()] = child.value.value  # pylint: disable=no-member # this is a fales positive
                    else:
                        settingsDict[child.getLocalAddress()] = child.value  # pylint: disable=no-member # this is a fales positive
            data = json.dumps(settingsDict)
            putResponse = self.vehicle.weConnect.session.put(url, data=data, allow_redirects=True)
            if putResponse.status_code != requests.codes['ok']:
                errorDict = putResponse.json()
                if errorDict is not None and 'error' in errorDict:
                    error = Error(localAddress='error', parent=self, fromDict=errorDict['error'])
                    if error is not None:
                        message = ''
                        if error.message.enabled and error.message.value is not None:
                            message += error.message.value
                        if error.info.enabled and error.info.value is not None:
                            message += ' - ' + error.info.value
                        if error.retry.enabled and error.retry.value is not None:
                            if error.retry.value:
                                message += ' - Please retry in a moment'
                            else:
                                message += ' - No retry possible'
                        raise SetterError(f'Could not set value ({message})')
                    else:
                        raise SetterError(f'Could not set value ({putResponse.status_code})')
                raise SetterError(f'Could not not set value ({putResponse.status_code})')
            responseDict = putResponse.json()
            if 'data' in responseDict and 'requestID' in responseDict['data']:
                if self.vehicle.requestTracker is not None:
                    self.vehicle.requestTracker.trackRequest(responseDict['data']['requestID'], Domain.ALL, 20, 120)



from typing import TYPE_CHECKING, Tuple
from datetime import datetime, timedelta
from threading import Timer
import logging

if TYPE_CHECKING:
    from weconnect.elements.vehicle import Vehicle

from weconnect.elements.generic_status import GenericStatus
from weconnect.domain import Domain


LOG = logging.getLogger("weconnect")


class RequestTracker:
    def __init__(self, vehicle: 'Vehicle') -> None:
        self.vehicle: 'Vehicle' = vehicle
        self.requests: dict[Domain, list[Tuple[str, datetime, datetime]]] = {}
        self.__timer = None

    def clear(self) -> None:
        self.requests.clear()
        if self.__timer is not None or self.__timer.is_alive():
            self.__timer.cancel()

    def trackRequest(self, id: str, domain: Domain, minTime: int, maxTime: int) -> None:
        minDate = datetime.now() + timedelta(seconds=minTime)
        maxDate = datetime.now() + timedelta(seconds=maxTime)
        LOG.debug('Track requests for id %s in %s at lesat until %s, at most until %s every 5 seconds', id, domain.value, minDate, maxDate)
        if domain not in self.requests:
            self.requests[domain] = [(id, minDate, maxDate)]
        else:
            self.requests[domain].append((id, minDate, maxDate))

        if self.__timer is None or not self.__timer.is_alive():
            self.__timer = Timer(5, self.update)
            self.__timer.daemon = True
            self.__timer.start()

    def update(self) -> None:  # noqa: C901
        LOG.debug('request tracking update status for %d domains', len(self.requests))
        self.vehicle.updateStatus(force=True, selective=self.requests)
        openRequests = []
        for domain, statuses in self.vehicle.domains.items():
            for status in statuses.values():
                if status.hasRequests():
                    openRequests.extend(status.requests.values())

        for domain, requests in list(self.requests.items()):
            for request in requests:
                id, minDate, maxDate = request

                if maxDate < datetime.now():
                    requests.remove(request)
                    LOG.debug('request tracking for id %s timed out', id)
                else:
                    if openRequests:
                        for openRequest in openRequests:
                            if openRequest.requestId.value == id:
                                if openRequest.status.value not in (GenericStatus.Request.Status.IN_PROGRESS,
                                                                    GenericStatus.Request.Status.QUEUED,
                                                                    GenericStatus.Request.Status.DELAYED):
                                    requests.remove(request)
                                    LOG.debug('request tracking for id %s finished with status %s', id, openRequest.status.value)
                                request = (id, datetime.now(), maxDate)
                    else:
                        if minDate < datetime.now():
                            requests.remove(request)
                            LOG.debug('request tracking for id %s removed, no requests seen until minimum time', id)
            if not requests:
                self.requests.pop(domain)

        if self.requests:  # Restart timer if there are still requests left
            self.__timer = Timer(5, self.update)
            self.__timer.daemon = True
            self.__timer.start()



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class PlugStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.plugConnectionState = AddressableAttribute(
            localAddress='plugConnectionState', parent=self, value=None, valueType=PlugStatus.PlugConnectionState)
        self.plugLockState = AddressableAttribute(
            localAddress='plugLockState', value=None, parent=self, valueType=PlugStatus.PlugLockState)
        self.externalPower = AddressableAttribute(
            localAddress='externalPower', value=None, parent=self, valueType=PlugStatus.ExternalPower)
        self.ledColor = AddressableAttribute(localAddress='ledColor', value=None, parent=self, valueType=PlugStatus.LedColor)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Plug status from dict')

        if 'value' in fromDict:
            self.plugConnectionState.fromDict(fromDict['value'], 'plugConnectionState')
            self.plugLockState.fromDict(fromDict['value'], 'plugLockState')
            self.externalPower.fromDict(fromDict['value'], 'externalPower')
            self.ledColor.fromDict(fromDict['value'], 'ledColor')
        else:
            self.plugConnectionState.enabled = False
            self.plugLockState.enabled = False
            self.externalPower.enabled = False
            self.ledColor.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['plugConnectionState', 'plugLockState', 'externalPower', 'ledColor']))

    def __str__(self):
        string = super().__str__()
        string += '\n\tPlug:'
        if self.plugConnectionState.enabled:
            string += f' {self.plugConnectionState.value.value}, '  # pylint: disable=no-member
        if self.plugLockState.enabled:
            string += f'{self.plugLockState.value.value}'  # pylint: disable=no-member
        if self.externalPower.enabled:
            string += f', External Power: {self.externalPower.value.value}'  # pylint: disable=no-member
        if self.ledColor.enabled:
            string += f', Led color: {self.ledColor.value.value}'  # pylint: disable=no-member
        return string

    class PlugConnectionState(Enum,):
        CONNECTED = 'connected'
        DISCONNECTED = 'disconnected'
        INVALID = 'invalid'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown unlock plug state'

    class PlugLockState(Enum,):
        LOCKED = 'locked'
        UNLOCKED = 'unlocked'
        INVALID = 'invalid'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown unlock plug state'

    class ExternalPower(Enum,):
        READY = 'ready'
        ACTIVE = 'active'
        UNAVAILABLE = 'unavailable'
        UNSUPPORTED = 'unsupported'
        UNKNOWN = 'unknown external power'

    class LedColor(Enum,):
        NONE = 'none'
        GREEN = 'green'
        RED = 'red'
        UNKNOWN = 'unknown plug led color'



import logging

from weconnect.addressable import AddressableLeaf, ChangeableAttribute
from weconnect.elements.generic_settings import GenericSettings
from weconnect.elements.enums import UnlockPlugState, MaximumChargeCurrent

LOG = logging.getLogger("weconnect")


class ChargingSettings(GenericSettings):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.maxChargeCurrentAC = ChangeableAttribute(
            localAddress='maxChargeCurrentAC', parent=self, value=None, valueType=MaximumChargeCurrent)
        self.autoUnlockPlugWhenCharged = ChangeableAttribute(localAddress='autoUnlockPlugWhenCharged', value=None,
                                                             parent=self, valueType=UnlockPlugState)
        self.autoUnlockPlugWhenChargedAC = ChangeableAttribute(localAddress='autoUnlockPlugWhenChargedAC', value=None,
                                                               parent=self, valueType=UnlockPlugState)
        self.targetSOC_pct = ChangeableAttribute(localAddress='targetSOC_pct', value=None, parent=self, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

        self.maxChargeCurrentAC.addObserver(self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                            priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.autoUnlockPlugWhenCharged.addObserver(self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                   priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.autoUnlockPlugWhenChargedAC.addObserver(self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                                     priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)
        self.targetSOC_pct.addObserver(self.valueChanged, AddressableLeaf.ObserverEvent.VALUE_CHANGED,
                                       priority=AddressableLeaf.ObserverPriority.INTERNAL_MID)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update Charging settings from dict')

        if 'value' in fromDict:
            self.maxChargeCurrentAC.fromDict(fromDict['value'], 'maxChargeCurrentAC')
            self.autoUnlockPlugWhenCharged.fromDict(fromDict['value'], 'autoUnlockPlugWhenCharged')
            self.autoUnlockPlugWhenChargedAC.fromDict(fromDict['value'], 'autoUnlockPlugWhenChargedAC')
            self.targetSOC_pct.fromDict(fromDict['value'], 'targetSOC_pct')
        else:
            self.maxChargeCurrentAC.enabled = False
            self.autoUnlockPlugWhenCharged.enabled = False
            self.autoUnlockPlugWhenChargedAC.enabled = False
            self.targetSOC_pct.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + [
                                                                'maxChargeCurrentAC',
                                                                'autoUnlockPlugWhenCharged',
                                                                'autoUnlockPlugWhenChargedAC',
                                                                'targetSOC_pct'
                                                            ]))

    def __str__(self):
        string = super().__str__()
        if self.maxChargeCurrentAC.enabled:
            string += f'\n\tMaximum Charge Current AC: {self.maxChargeCurrentAC.value.value}'  # pylint: disable=no-member # this is a fales positive
        if self.autoUnlockPlugWhenCharged.enabled:
            string += f'\n\tAuto Unlock When Charged: {self.autoUnlockPlugWhenCharged.value.value}'  # pylint: disable=no-member # this is a fales positive
        if self.autoUnlockPlugWhenChargedAC.enabled:
            string += f'\n\tAuto Unlock When Charged AC: {self.autoUnlockPlugWhenChargedAC.value.value}'  # pylint: disable=no-member # this is a fales positive
        if self.targetSOC_pct.enabled:
            string += f'\n\tTarget SoC: {self.targetSOC_pct.value} %'
        return string



from enum import Enum
import base64
import io
import logging

from weconnect.addressable import AddressableAttribute, AddressableObject, AddressableDict
from weconnect.elements.generic_status import GenericStatus

SUPPORT_IMAGES = False
try:
    from PIL import Image  # type: ignore
    SUPPORT_IMAGES = True
except ImportError:
    pass

LOG = logging.getLogger("weconnect")


class WarningLightsStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.mileage_km = AddressableAttribute(localAddress='mileage_km', value=None, parent=self, valueType=int)
        self.warningLights = AddressableDict(localAddress='warningLights', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update warning lights status from dict')

        if 'value' in fromDict:
            if 'mileage_km' in fromDict['value']:
                mileage_km = int(fromDict['value']['mileage_km'])
                if self.fixAPI and mileage_km == 0x7FFFFFFF:
                    mileage_km = None
                    LOG.info('%s: Attribute mileage_km was error value 0x7FFFFFFF. Setting error state instead'
                             ' of 2147483647 km.', self.getGlobalAddress())
                self.mileage_km.setValueWithCarTime(mileage_km, lastUpdateFromCar=None, fromServer=True)
            else:
                self.mileage_km.enabled = False

            if 'warningLights' in fromDict['value'] and fromDict['value']['warningLights'] is not None:
                for warningLightDict in fromDict['value']['warningLights']:
                    if 'messageId' in warningLightDict:
                        if warningLightDict['messageId'] in self.warningLights:
                            self.warningLights[warningLightDict['messageId']].update(fromDict=warningLightDict)
                        else:
                            self.warningLights[warningLightDict['messageId']] = WarningLightsStatus.WarningLight(fromDict=warningLightDict,
                                                                                                                 parent=self.warningLights)
                for warningLightMessageId in [warningLightMessageId for warningLightMessageId in self.warningLights.keys()
                                              if warningLightMessageId not in [warningLight['messageId'] for warningLight in fromDict['value']['warningLights']
                                                                               if 'messageId' in warningLight]]:
                    del self.warningLights[warningLightMessageId]
            else:
                self.warningLights.clear()
                self.warningLights.enabled = False
        else:
            self.mileage_km.enabled = False
            self.warningLights.clear()
            self.warningLights.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes
                                                            + [
                                                                'mileage_km',
                                                                'warningLights',
                                                            ]))

    def __str__(self):
        string = super().__str__()
        if self.mileage_km.enabled:
            string += f'\n\tCurrent milage: {self.mileage_km.value} km'
        if self.warningLights.enabled and len(self.warningLights) > 0:
            string += f'\n\tWarning Lights: {len(self.warningLights)} items'
            for warningLight in self.warningLights.values():
                string += '\n' + ''.join(['\t\t' + line for line in str(warningLight).splitlines(True)])
        return string

    class WarningLight(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.text = AddressableAttribute(localAddress='text', parent=self, value=None, valueType=str)
            self.category = AddressableAttribute(localAddress='category', parent=self, value=None, valueType=WarningLightsStatus.WarningLight.Category)
            self.priority = AddressableAttribute(localAddress='priority', parent=self, value=None, valueType=int)
            self.icon = AddressableAttribute(localAddress='icon', parent=self, value=None, valueType=Image.Image)
            self.iconName = AddressableAttribute(localAddress='iconName', parent=self, value=None, valueType=str)
            self.messageId = AddressableAttribute(localAddress='messageId', parent=self, value=None, valueType=str)
            self.serviceLead = AddressableAttribute(localAddress='serviceLead', parent=self, value=None, valueType=bool)
            self.customerRelevance = AddressableAttribute(localAddress='customerRelevance', parent=self, value=None, valueType=bool)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update warning light from dict')

            if 'messageId' in fromDict:
                self.messageId.fromDict(fromDict, 'messageId')
                self.localAddress = self.messageId.value
            else:
                LOG.error('Warning light is missing messageId attribute')

            self.text.fromDict(fromDict, 'text')
            self.category.fromDict(fromDict, 'category')
            self.priority.fromDict(fromDict, 'priority')
            if SUPPORT_IMAGES:
                if 'icon' in fromDict and fromDict['icon'] is not None:
                    prefix = 'data:image/png;base64,'
                    if fromDict['icon'].startswith(prefix):
                        img = fromDict['icon'][len(prefix):]
                        img = base64.b64decode(img)
                        img = Image.open(io.BytesIO(img))
                        self.icon.setValueWithCarTime(img, lastUpdateFromCar=None, fromServer=True)
                    else:
                        LOG.error('%s: warning light icon is not a base64 encoded png', self.getGlobalAddress())
                        self.icon.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
                        self.icon.enabled = False
                else:
                    self.icon.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
                    self.icon.enabled = False

                self.iconName.fromDict(fromDict, 'iconName')
            self.serviceLead.fromDict(fromDict, 'serviceLead')
            self.customerRelevance.fromDict(fromDict, 'customerRelevance')

            for key, value in {key: value for key, value in fromDict.items() if key not in ['messageId', 'category', 'priority', 'icon', 'iconName',
                                                                                            'serviceLead', 'customerRelevance', 'text']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            returnStr = f'{self.messageId.value}: {self.text.value}'  # pylint: disable=no-member
            if self.category.enabled:
                returnStr += f'\n\tCategory: {self.category.value.value}'
            if self.category.enabled:
                returnStr += f'\n\tPriority: {self.priority.value}'
            if self.customerRelevance.enabled:
                returnStr += f'\n\tCustomer Relevance: {self.customerRelevance.value}'
            if self.serviceLead.enabled:
                returnStr += f'\n\tService Lead: {self.serviceLead.value}'
            return returnStr

        class Category(Enum,):
            LIGHTING = 'LIGHTING'
            UNKNOWN = 'unknown category'



from __future__ import annotations
from typing import Dict, List, Set, Any, Type, Optional, cast, TYPE_CHECKING
import os
from enum import Enum
from datetime import datetime, timedelta
import base64
import io
import logging

from weconnect.elements.generic_settings import GenericSettings
from weconnect.elements.generic_status import GenericStatus

from requests import exceptions, codes

from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableDict, AddressableList
if TYPE_CHECKING:
    from weconnect.weconnect import WeConnect
from weconnect.elements.generic_capability import GenericCapability
from weconnect.elements.generic_request_status import GenericRequestStatus
from weconnect.elements.controls import Controls
from weconnect.elements.access_status import AccessStatus
from weconnect.elements.battery_status import BatteryStatus
from weconnect.elements.lv_battery_status import LVBatteryStatus
from weconnect.elements.battery_support_status import BatterySupportStatus
from weconnect.elements.capability_status import CapabilityStatus
from weconnect.elements.charging_status import ChargingStatus
from weconnect.elements.charging_settings import ChargingSettings
from weconnect.elements.charge_mode import ChargeMode
from weconnect.elements.climatization_status import ClimatizationStatus
from weconnect.elements.climatization_settings import ClimatizationSettings
from weconnect.elements.climatization_timer import ClimatizationTimer
from weconnect.elements.departure_timers_status import DepartureTimersStatus
from weconnect.elements.lights_status import LightsStatus
from weconnect.elements.maintenance_status import MaintenanceStatus
from weconnect.elements.warning_lights_status import WarningLightsStatus
from weconnect.elements.parking_position import ParkingPosition
from weconnect.elements.plug_status import PlugStatus
from weconnect.elements.range_status import RangeStatus
from weconnect.elements.window_heating_status import WindowHeatingStatus
from weconnect.elements.odometer_measurement import OdometerMeasurement
from weconnect.elements.range_measurements import RangeMeasurements
from weconnect.elements.readiness_status import ReadinessStatus
from weconnect.elements.charging_profiles import ChargingProfiles
from weconnect.errors import APICompatibilityError, RetrievalError, APIError
from weconnect.util import toBool
from weconnect.weconnect_errors import ErrorEventType
from weconnect.domain import Domain
from weconnect.elements.error import Error

from weconnect.elements.helpers.request_tracker import RequestTracker

SUPPORT_IMAGES = False
try:
    from PIL import Image, ImageDraw  # type: ignore
    SUPPORT_IMAGES = True
except ImportError:
    pass

LOG: logging.Logger = logging.getLogger("weconnect")


class DomainDict(AddressableDict):
    def __init__(self, **kwargs):
        self.error: Error = Error(localAddress='error', parent=self)
        super(DomainDict, self).__init__(**kwargs)

    def updateError(self, fromDict: Dict[str, Any]):
        if 'error' in fromDict:
            self.error.update(fromDict['error'])
        else:
            self.error.reset()

    def hasError(self) -> bool:
        return self.error.enabled


class Vehicle(AddressableObject):  # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        weConnect: WeConnect,
        vin: str,
        parent: AddressableDict[str, Vehicle],
        fromDict: Dict[str, Any],
        fixAPI: bool = True,
        updateCapabilities: bool = True,
        updatePictures: bool = True,
        selective: Optional[list[Domain]] = None,
        enableTracker: bool = False
    ) -> None:
        self.weConnect: WeConnect = weConnect
        super().__init__(localAddress=vin, parent=parent)
        self.vin: AddressableAttribute[str] = AddressableAttribute(localAddress='vin', parent=self, value=None, valueType=str)
        self.role: AddressableAttribute[Vehicle.User.Role] = AddressableAttribute(localAddress='role', parent=self, value=None, valueType=Vehicle.User.Role)
        self.enrollmentStatus: AddressableAttribute[Vehicle.User.EnrollmentStatus] = AddressableAttribute(localAddress='enrollmentStatus', parent=self,
                                                                                                          value=None,
                                                                                                          valueType=Vehicle.User.EnrollmentStatus)
        self.userRoleStatus: AddressableAttribute[Vehicle.User.UserRoleStatus] = AddressableAttribute(localAddress='userRoleStatus', parent=self,
                                                                                                      value=None,
                                                                                                      valueType=Vehicle.User.UserRoleStatus)
        self.model: AddressableAttribute[str] = AddressableAttribute(localAddress='model', parent=self, value=None, valueType=str)
        self.devicePlatform: AddressableAttribute[Vehicle.DevicePlatform] = AddressableAttribute(localAddress='devicePlatform', parent=self,
                                                                                                 value=None,
                                                                                                 valueType=Vehicle.DevicePlatform)
        self.nickname: AddressableAttribute[str] = AddressableAttribute(localAddress='nickname', parent=self, value=None, valueType=str)
        self.brandCode: AddressableAttribute[str] = AddressableAttribute(localAddress='brandCode', parent=self, value=None, valueType=Vehicle.BrandCode)
        self.capabilities: AddressableDict[str, GenericCapability] = AddressableDict(localAddress='capabilities', parent=self)
        self.domains: AddressableDict[str, DomainDict[str, GenericStatus]] = AddressableDict(localAddress='domains', parent=self)
        self.images: AddressableAttribute[Dict[str, str]] = AddressableAttribute(localAddress='images', parent=self, value=None, valueType=dict)
        self.tags: AddressableAttribute[List[str]] = AddressableAttribute(localAddress='tags', parent=self, value=None, valueType=list)
        self.coUsers: AddressableList[Vehicle.User] = AddressableList(localAddress='coUsers', parent=self)
        self.controls: Controls = Controls(localAddress='controls', vehicle=self, parent=self)
        self.fixAPI: bool = fixAPI

        if SUPPORT_IMAGES:
            self.__carImages: Dict[str, Image.Image] = {}
            self.__badges: Dict[Vehicle.Badge, Image.Image] = {}
            self.pictures: AddressableDict[str, Image.Image] = AddressableDict(localAddress='pictures', parent=self)

        self.requestTracker: Optional[RequestTracker] = None
        if enableTracker:
            self.requestTracker = RequestTracker(self)

        self.update(fromDict, updateCapabilities=updateCapabilities, updatePictures=updatePictures, selective=selective)

    def enableTracker(self) -> None:
        if self.requestTracker is None:
            self.requestTracker = RequestTracker(self)

    def disableTracker(self) -> None:
        self.requestTracker.clear()
        self.requestTracker = None

    def statusExists(self, domain: str, status: str) -> bool:
        if domain in self.domains and status in self.domains[domain]:
            return True
        return False

    def update(  # noqa: C901  # pylint: disable=too-many-branches
        self,
        fromDict: Dict[str, Any] = None,
        updateCapabilities: bool = True,
        updatePictures: bool = True,
        force: bool = False,
        selective: Optional[list[Domain]] = None
    ) -> None:
        if fromDict is not None:
            LOG.debug('Create /update vehicle')

            self.vin.fromDict(fromDict, 'vin')
            self.role.fromDict(fromDict, 'role')
            self.enrollmentStatus.fromDict(fromDict, 'enrollmentStatus')
            self.userRoleStatus.fromDict(fromDict, 'userRoleStatus')
            self.model.fromDict(fromDict, 'model')
            self.devicePlatform.fromDict(fromDict, 'devicePlatform')
            self.nickname.fromDict(fromDict, 'nickname')
            self.brandCode.fromDict(fromDict, 'brandCode')

            if updateCapabilities and 'capabilities' in fromDict and fromDict['capabilities'] is not None:
                for capDict in fromDict['capabilities']:
                    if 'id' in capDict:
                        if capDict['id'] in self.capabilities:
                            self.capabilities[capDict['id']].update(fromDict=capDict)
                        else:
                            self.capabilities[capDict['id']] = GenericCapability(
                                capabilityId=capDict['id'], parent=self.capabilities, fromDict=capDict,
                                fixAPI=self.fixAPI)
                for capabilityId in [capabilityId for capabilityId in self.capabilities.keys()
                                     if capabilityId not in [capability['id']
                                     for capability in fromDict['capabilities'] if 'id' in capability]]:
                    del self.capabilities[capabilityId]
            else:
                self.capabilities.clear()
                self.capabilities.enabled = False

            if 'images' in fromDict:
                self.images.setValueWithCarTime(fromDict['images'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.images.enabled = False

            if 'tags' in fromDict:
                self.tags.setValueWithCarTime(fromDict['tags'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.tags.enabled = False

            if 'coUsers' in fromDict and fromDict['coUsers'] is not None:
                for user in fromDict['coUsers']:
                    if 'id' in user:
                        usersWithId = [x for x in self.coUsers if x.id.value == user['id']]
                        if len(usersWithId) > 0:
                            usersWithId[0].update(fromDict=user)
                        else:
                            self.coUsers.append(Vehicle.User(localAddress=str(len(self.coUsers)), parent=self.coUsers, fromDict=user))
                    else:
                        raise APICompatibilityError('User is missing id field')
                # Remove all users that are not in list anymore
                for user in [user for user in self.coUsers if user.id.value not in [x['id'] for x in fromDict['coUsers']]]:
                    self.coUsers.remove(user)
            else:
                self.coUsers.enabled = False
                self.coUsers.clear()

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['vin',
                                              'role',
                                              'enrollmentStatus',
                                              'userRoleStatus',
                                              'model',
                                              'devicePlatform',
                                              'nickname',
                                              'brandCode',
                                              'capabilities',
                                              'images',
                                              'tags',
                                              'coUsers']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        self.updateStatus(updateCapabilities=updateCapabilities, force=force, selective=selective)
        if SUPPORT_IMAGES and updatePictures:
            for badge in Vehicle.Badge:
                badgeImg: Image = Image.open(f'{os.path.dirname(__file__)}/../badges/{badge.value}.png')
                badgeImg.thumbnail((100, 100))
                self.__badges[badge] = badgeImg

            self.updatePictures()

    def updateStatus(self, updateCapabilities: bool = True, force: bool = False,  # noqa: C901 # pylint: disable=too-many-branches
                     selective: Optional[list[Domain]] = None):
        jobKeyClassMap: Dict[Domain, Dict[str, Type[GenericStatus]]] = {
            Domain.ACCESS: {
                'accessStatus': AccessStatus
            },
            Domain.AUTOMATION: {
                'climatisationTimer': ClimatizationTimer,
                'climatisationTimersRequestStatus': GenericRequestStatus,
                'chargingProfiles': ChargingProfiles,
            },
            Domain.USER_CAPABILITIES: {
                'capabilitiesStatus': CapabilityStatus,
            },
            Domain.CHARGING: {
                'batteryStatus': BatteryStatus,
                'chargingStatus': ChargingStatus,
                'chargingSettings': ChargingSettings,
                'chargeMode': ChargeMode,
                'plugStatus': PlugStatus,
                'chargingRequestStatus': GenericRequestStatus,
                'chargingSettingsRequestStatus': GenericRequestStatus,
                'chargingCareSettings': GenericSettings,
            },
            Domain.CHARGING_PROFILES: {
                'chargingProfilesStatus': ChargingProfiles,
            },
            Domain.CLIMATISATION: {
                'climatisationStatus': ClimatizationStatus,
                'climatisationSettings': ClimatizationSettings,
                'windowHeatingStatus': WindowHeatingStatus,
                'climatisationRequestStatus': GenericRequestStatus,
                'climatisationSettingsRequestStatus': GenericRequestStatus,
            },
            Domain.DEPARTURE_TIMERS: {
                'departureTimersStatus': DepartureTimersStatus,
            },
            Domain.FUEL_STATUS: {
                'rangeStatus': RangeStatus,
            },
            Domain.VEHICLE_LIGHTS: {
                'lightsStatus': LightsStatus,
            },
            Domain.LV_BATTERY: {
                'lvBatteryStatus': LVBatteryStatus,
            },
            Domain.READINESS: {
                'readinessStatus': ReadinessStatus,
                'readinessBatterySupportStatus': GenericStatus,
            },
            Domain.VEHICLE_HEALTH_INSPECTION: {
                'maintenanceStatus': MaintenanceStatus,
            },
            Domain.VEHICLE_HEALTH_WARNINGS: {
                'warningLights': WarningLightsStatus,
            },
            Domain.OIL_LEVEL: {
                'oilLevelStatus': GenericStatus,
            },
            Domain.MEASUREMENTS: {
                'rangeStatus': RangeMeasurements,
                'odometerStatus': OdometerMeasurement,
                'oilLevelStatus': GenericStatus,
                'measurements': GenericStatus,
            },
            Domain.BATTERY_SUPPORT: {
                'batterySupportStatus': BatterySupportStatus,
            }
        }
        if self.vin.value is None:
            raise APIError('')
        if selective is None:
            jobs = [domain.value for domain in Domain if domain != Domain.ALL and domain != Domain.ALL_CAPABLE and domain != Domain.PARKING]
        elif Domain.ALL_CAPABLE in selective:
            if self.capabilities:
                jobs = []
                for dom in [domain for domain in Domain if domain != Domain.ALL and domain != Domain.ALL_CAPABLE and domain != Domain.PARKING]:
                    if dom.value in self.capabilities and self.capabilities[dom.value].enabled and not self.capabilities[dom.value].status.enabled:
                        jobs.append(dom.value)
                if updateCapabilities:
                    jobs.append(Domain.USER_CAPABILITIES.value)
            else:
                jobs = ['all']
        elif Domain.ALL in selective:
            jobs = ['all']
        else:
            jobs = [domain.value for domain in selective]
        url: str = 'https://mobileapi.apps.emea.vwapps.io/vehicles/' + self.vin.value + '/selectivestatus?jobs=' + ','.join(jobs)
        data: Optional[Dict[str, Any]] = self.weConnect.fetchData(url, force)
        if data is not None:
            for domain, keyClassMap in jobKeyClassMap.items():
                if not updateCapabilities and domain == Domain.USER_CAPABILITIES:
                    continue
                if domain.value in data:
                    if domain.value not in self.domains:
                        self.domains[domain.value] = DomainDict(localAddress=domain.value, parent=self.domains)
                    for key, className in keyClassMap.items():
                        if key in data[domain.value]:
                            if key in self.domains[domain.value]:
                                LOG.debug('Status %s exists, updating it', key)
                                self.domains[domain.value][key].update(fromDict=data[domain.value][key])
                            else:
                                LOG.debug('Status %s does not exist, creating it', key)
                                self.domains[domain.value][key] = className(vehicle=self, parent=self.domains[domain.value], statusId=key,
                                                                            fromDict=data[domain.value][key], fixAPI=self.fixAPI)
                    if 'error' in data[domain.value]:
                        self.domains[domain.value].updateError(data[domain.value])

                    # check that there is no additional status than the configured ones, except for "target" that we merge into
                    # the known ones
                    for key, value in {key: value for key, value in data[domain.value].items()
                                       if key not in list(keyClassMap.keys()) and key not in ['error']}.items():
                        LOG.warning('%s: Unknown attribute %s with value %s in domain %s', self.getGlobalAddress(), key, value, domain.value)
            # check that there is no additional domain than the configured ones
            for key, value in {key: value for key, value in data.items() if key not in list([domain.value for domain in jobKeyClassMap.keys()])}.items():
                LOG.warning('%s: Unknown domain %s with value %s', self.getGlobalAddress(), key, value)

        # Controls
        self.controls.update()

        if (selective is None or any(x in selective for x in [Domain.ALL, Domain.ALL_CAPABLE, Domain.PARKING])) \
                and (not updateCapabilities or ('parkingPosition' in self.capabilities and self.capabilities['parkingPosition'].status.value is None)):
            url = 'https://mobileapi.apps.emea.vwapps.io/vehicles/' + self.vin.value + '/parkingposition'
            data = self.weConnect.fetchData(url, force, allowEmpty=True, allowHttpError=True, allowedErrors=[codes['not_found'],
                                                                                                             codes['no_content'],
                                                                                                             codes['bad_gateway'],
                                                                                                             codes['forbidden']])

            if data is not None:
                if 'parking' not in self.domains:
                    self.domains['parking'] = DomainDict(localAddress='parking', parent=self)
                if 'parkingPosition' in self.domains['parking']:
                    self.domains['parking']['parkingPosition'].update(fromDict=data)
                else:
                    self.domains['parking']['parkingPosition'] = ParkingPosition(vehicle=self,
                                                                                 parent=self.domains['parking'],
                                                                                 statusId='parkingPosition',
                                                                                 fromDict=data)
            else:
                if self.statusExists('parking', 'parkingPosition'):
                    parkingPosition: ParkingPosition = cast(ParkingPosition, self.domains['parking']['parkingPosition'])
                    parkingPosition.latitude.enabled = False
                    parkingPosition.longitude.enabled = False
                    parkingPosition.carCapturedTimestamp.setValueWithCarTime(None, fromServer=True)
                    parkingPosition.carCapturedTimestamp.enabled = False
                    parkingPosition.enabled = False

    def updatePictures(self) -> None:  # noqa: C901
        if not SUPPORT_IMAGES:
            return
        url: str = f'https://vehicle-images-service.apps.emea.vwapps.io/v2/vehicle-images/{self.vin.value}?resolution=2x'
        data = self.weConnect.fetchData(url, allowHttpError=True)
        if data is not None and 'data' in data:  # pylint: disable=too-many-nested-blocks
            for image in data['data']:
                img = None
                cacheDate = None
                imageurl: str = image['url']
                if self.weConnect.maxAgePictures is not None and self.weConnect.cache is not None and imageurl in self.weConnect.cache:
                    img, cacheDateString = self.weConnect.cache[imageurl]
                    img = base64.b64decode(img)
                    img = Image.open(io.BytesIO(img))
                    cacheDate = datetime.fromisoformat(cacheDateString)
                if img is None or self.weConnect.maxAgePictures is None \
                        or (cacheDate is not None and cacheDate < (datetime.utcnow() - timedelta(seconds=self.weConnect.maxAgePictures))):
                    try:
                        imageDownloadResponse = self.weConnect.session.get(imageurl, stream=True)
                        self.weConnect.recordElapsed(imageDownloadResponse.elapsed)
                        if imageDownloadResponse.status_code == codes['ok']:
                            img = Image.open(imageDownloadResponse.raw)
                            if self.weConnect.cache is not None:
                                buffered = io.BytesIO()
                                img.save(buffered, format="PNG")
                                imgStr = base64.b64encode(buffered.getvalue()).decode("utf-8")
                                self.weConnect.cache[imageurl] = (imgStr, str(datetime.utcnow()))
                        elif imageDownloadResponse.status_code == codes['unauthorized']:
                            LOG.info('Server asks for new authorization')
                            self.weConnect.login()
                            imageDownloadResponse = self.weConnect.session.get(imageurl, stream=True)
                            self.weConnect.recordElapsed(imageDownloadResponse.elapsed)
                            if imageDownloadResponse.status_code == codes['ok']:
                                img = Image.open(imageDownloadResponse.raw)
                                if self.weConnect.cache is not None:
                                    buffered = io.BytesIO()
                                    img.save(buffered, format="PNG")
                                    imgStr = base64.b64encode(buffered.getvalue()).decode("utf-8")
                                    self.weConnect.cache[imageurl] = (imgStr, str(datetime.utcnow()))
                            else:
                                self.weConnect.notifyError(self, ErrorEventType.HTTP, str(imageDownloadResponse.status_code),
                                                           'Could not fetch vehicle image due to server error')
                                raise RetrievalError('Could not retrieve vehicle image even after re-authorization.'
                                                     f' Status Code was: {imageDownloadResponse.status_code}')
                            self.weConnect.notifyError(self, ErrorEventType.HTTP, str(imageDownloadResponse.status_code),
                                                       'Could not fetch vehicle image due to server error')
                            raise RetrievalError(f'Could not retrieve vehicle image. Status Code was: {imageDownloadResponse.status_code}')
                        else:
                            LOG.warning('Failed downloading picture %s with status code %d will try again in next update', image['id'],
                                        imageDownloadResponse.status_code)
                    except exceptions.ConnectionError as connectionError:
                        self.weConnect.notifyError(self, ErrorEventType.CONNECTION, 'connection',
                                                   'Could not fetch vehicle image due to connection problem')
                        raise RetrievalError from connectionError
                    except exceptions.ChunkedEncodingError as chunkedEncodingError:
                        self.weConnect.notifyError(self, ErrorEventType.CONNECTION, 'chunked encoding error',
                                                   'Could not refresh token due to connection problem with chunked encoding')
                        raise RetrievalError from chunkedEncodingError
                    except exceptions.ReadTimeout as timeoutError:
                        self.weConnect.notifyError(self, ErrorEventType.TIMEOUT, 'timeout', 'Could not fetch vehicle image due to timeout')
                        raise RetrievalError from timeoutError
                    except exceptions.RetryError as retryError:
                        raise RetrievalError from retryError

                if img is not None:
                    self.__carImages[image['id']] = img
                    if image['id'] == 'car_34view':
                        if 'car' in self.pictures:
                            self.pictures['car'].setValueWithCarTime(self.__carImages['car_34view'], lastUpdateFromCar=None, fromServer=True)
                        else:
                            self.pictures['car'] = AddressableAttribute(localAddress='car', parent=self.pictures, value=self.__carImages['car_34view'],
                                                                        valueType=Image.Image)

            self.updateStatusPicture()

    def updateStatusPicture(self) -> None:  # noqa: C901
        if not SUPPORT_IMAGES:
            return
        if 'car_birdview' in self.__carImages:
            img: Image = self.__carImages['car_birdview']

            badges: Set[Vehicle.Badge] = set()

            doorNameMap: Dict[str, str] = {'frontLeft': 'door_left_front',
                                           'frontRight': 'door_right_front',
                                           'rearLeft': 'door_left_back',
                                           'rearRight': 'door_right_back'}
            windowNameMap: Dict[str, str] = {'frontLeft': 'window_left_front',
                                             'frontRight': 'window_right_front',
                                             'rearLeft': 'window_left_back',
                                             'rearRight': 'window_right_back',
                                             'sunRoof': 'sunroof'}
            if 'access' in self.domains and 'accessStatus' in self.domains['access'] and not self.domains['access']['accessStatus'].error.enabled:
                accessStatus: AccessStatus = cast(AccessStatus, self.domains['access']['accessStatus'])

                if accessStatus.overallStatus.enabled:
                    if accessStatus.overallStatus.value == AccessStatus.OverallState.SAFE:
                        badges.add(Vehicle.Badge.LOCKED)
                    elif accessStatus.overallStatus.value == AccessStatus.OverallState.UNSAFE:
                        badges.add(Vehicle.Badge.UNLOCKED)
                    else:
                        badges.add(Vehicle.Badge.WARNING)

                for name, door in accessStatus.doors.items():
                    name = doorNameMap.get(name, name)
                    doorImageName: Optional[str] = None

                    if door.openState.value == AccessStatus.Door.OpenState.OPEN:
                        doorImageName = f'{name}_overlay'
                    elif door.openState.value == AccessStatus.Door.OpenState.CLOSED:
                        doorImageName = name
                    elif door.openState.value in (AccessStatus.Door.OpenState.INVALID, AccessStatus.Door.OpenState.UNKNOWN):
                        doorImageName = name
                        badges.add(Vehicle.Badge.WARNING)

                    if doorImageName is not None and doorImageName in self.__carImages:
                        doorImage = self.__carImages[doorImageName].convert("RGBA")
                        img.paste(doorImage, (0, 0), doorImage)

                for name, window in accessStatus.windows.items():
                    name = windowNameMap.get(name, name)
                    windowImageName: Optional[str] = None

                    if window.openState.value == AccessStatus.Window.OpenState.OPEN:
                        windowImageName = f'{name}_overlay'
                    elif window.openState.value == AccessStatus.Window.OpenState.CLOSED:
                        windowImageName = name
                    elif window.openState.value in (AccessStatus.Window.OpenState.INVALID, AccessStatus.Window.OpenState.UNKNOWN):
                        windowImageName = name
                        badges.add(Vehicle.Badge.WARNING)

                    if windowImageName is not None and windowImageName in self.__carImages:
                        windowImage = self.__carImages[windowImageName].convert("RGBA")
                        img.paste(windowImage, (0, 0), windowImage)
            else:
                for name in doorNameMap.values():
                    if name in self.__carImages:
                        doorImage = self.__carImages[name].convert("RGBA")
                        img.paste(doorImage, (0, 0), doorImage)
                for name in windowNameMap.values():
                    if name != 'sunroof' and name in self.__carImages:
                        windowImage = self.__carImages[name].convert("RGBA")
                        img.paste(windowImage, (0, 0), windowImage)

            if 'vehicleLights' in self.domains and 'lightsStatus' in self.domains['vehicleLights']:
                lightsStatus: LightsStatus = cast(LightsStatus, self.domains['vehicleLights']['lightsStatus'])
                for name, light in lightsStatus.lights.items():
                    lightNameMap = {'frontLeft': 'door_left_front',
                                    'frontRight': 'door_right_front',
                                    'rearLeft': 'door_left_back',
                                    'rearRight': 'door_right_back'}
                    name = lightNameMap.get(name, name)
                    lightImageName: Optional[str] = None

                    if light.status.value == LightsStatus.Light.LightState.ON:
                        lightImageName = f'light_{name}'
                        if lightImageName in self.__carImages:
                            lightImage = self.__carImages[lightImageName].convert("RGBA")
                            img.paste(lightImage, (0, 0), lightImage)

            if 'charging' in self.domains and 'chargingStatus' in self.domains['charging']:
                chargingStatus: ChargingStatus = cast(ChargingStatus, self.domains['charging']['chargingStatus'])
                if chargingStatus.chargingState.value in (ChargingStatus.ChargingState.CHARGING,
                                                          ChargingStatus.ChargingState.CHARGE_PURPOSE_REACHED_CONSERVATION,
                                                          ChargingStatus.ChargingState.CONSERVATION):
                    badges.add(Vehicle.Badge.CHARGING)
                elif chargingStatus.chargingState.value == ChargingStatus.ChargingState.ERROR:
                    badges.add(Vehicle.Badge.WARNING)

            if 'charging' in self.domains and 'plugStatus' in self.domains['charging']:
                plugStatus: PlugStatus = cast(PlugStatus, self.domains['charging']['plugStatus'])
                if plugStatus.plugConnectionState.value == PlugStatus.PlugConnectionState.CONNECTED:
                    badges.add(Vehicle.Badge.CONNECTED)

            if 'climatisation' in self.domains and 'climatisationStatus' in self.domains['climatisation']:
                climatisationStatus: ClimatizationStatus = cast(ClimatizationStatus, self.domains['climatisation']['climatisationStatus'])
                if climatisationStatus.climatisationState.value == ClimatizationStatus.ClimatizationState.COOLING:
                    badges.add(Vehicle.Badge.COOLING)
                elif climatisationStatus.climatisationState.value == ClimatizationStatus.ClimatizationState.HEATING:
                    badges.add(Vehicle.Badge.HEATING)
                elif climatisationStatus.climatisationState.value == ClimatizationStatus.ClimatizationState.VENTILATION:
                    badges.add(Vehicle.Badge.VENTILATING)

            if 'parking' in self.domains and 'parkingPosition' in self.domains['parking']:
                parkingPosition: ParkingPosition = cast(ParkingPosition, self.domains['parking']['parkingPosition'])
                if parkingPosition.latitude.enabled and parkingPosition.latitude.value is not None:
                    badges.add(Vehicle.Badge.PARKING)

            self.__carImages['status'] = img

            imgWithBadges = img.copy()
            badgeoffset = 0
            for badge in badges:
                badgeImage = self.__badges[badge].convert("RGBA")
                imgWithBadges.paste(badgeImage, (0, badgeoffset), badgeImage)
                badgeoffset += 110

            warningLightoffset = 0
            imgWidth, _ = imgWithBadges.size
            if 'vehicleHealthWarnings' in self.domains and 'warningLights' in self.domains['vehicleHealthWarnings']:
                warningLightsStatus = self.domains['vehicleHealthWarnings']['warningLights']
                if warningLightsStatus.warningLights.enabled:
                    for warningLight in warningLightsStatus.warningLights.values():
                        if warningLight.icon.enabled:
                            draw = ImageDraw.Draw(imgWithBadges)
                            draw.ellipse(((imgWidth - 100), warningLightoffset, (imgWidth - 1), (warningLightoffset + 100)), fill=(0, 0, 0, 200))
                            lightImage = warningLight.icon.value
                            lightImage = lightImage.resize((64, 64), Image.ANTIALIAS)
                            imgWithBadges.paste(lightImage, ((imgWidth - 82), warningLightoffset + 18), lightImage)
                            warningLightoffset += 110

            self.__carImages['statusWithBadge'] = imgWithBadges

            # Car with badges
            if 'car_34view' in self.__carImages:
                carWithBadges = self.__carImages['car_34view'].copy()

                badgeoffset = 0
                for badge in badges:
                    badgeImage = self.__badges[badge].convert("RGBA")
                    carWithBadges.paste(badgeImage, (0, badgeoffset), badgeImage)
                    badgeoffset += 110

                warningLightoffset = 0
                imgWidth, _ = carWithBadges.size
                if 'vehicleHealthWarnings' in self.domains and 'warningLights' in self.domains['vehicleHealthWarnings']:
                    warningLightsStatus = self.domains['vehicleHealthWarnings']['warningLights']
                    if warningLightsStatus.warningLights.enabled:
                        for warningLight in warningLightsStatus.warningLights.values():
                            if warningLight.icon.enabled:
                                draw = ImageDraw.Draw(carWithBadges)
                                draw.ellipse(((imgWidth - 100), warningLightoffset, (imgWidth - 1), (warningLightoffset + 100)), fill=(0, 0, 0, 200))
                                lightImage = warningLight.icon.value
                                lightImage = lightImage.resize((64, 64), Image.ANTIALIAS)
                                carWithBadges.paste(lightImage, ((imgWidth - 82), warningLightoffset + 18), lightImage)
                                warningLightoffset += 110

                self.__carImages['carWithBadge'] = carWithBadges

        else:
            LOG.info('Could not update status picture as birdview image could not be retrieved')
            if 'status' in self.__carImages:
                self.__carImages.pop("status")

        if 'status' in self.pictures:
            if 'status' in self.__carImages:
                self.pictures['status'].setValueWithCarTime(self.__carImages['status'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.pictures['status'].setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
                self.pictures['status'].enabled = False
        else:
            if 'status' in self.__carImages:
                self.pictures['status'] = AddressableAttribute(localAddress='status', parent=self.pictures, value=self.__carImages['status'],
                                                               valueType=Image.Image)

        if 'statusWithBadge' in self.pictures:
            if 'statusWithBadge' in self.__carImages:
                self.pictures['statusWithBadge'].setValueWithCarTime(self.__carImages['statusWithBadge'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.pictures['statusWithBadge'].setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
                self.pictures['statusWithBadge'].enabled = False
        else:
            if 'statusWithBadge' in self.__carImages:
                self.pictures['statusWithBadge'] = AddressableAttribute(localAddress='statusWithBadge', parent=self.pictures,
                                                                        value=self.__carImages['statusWithBadge'],
                                                                        valueType=Image.Image)

        if 'carWithBadge' in self.pictures:
            if 'carWithBadge' in self.__carImages:
                self.pictures['carWithBadge'].setValueWithCarTime(self.__carImages['carWithBadge'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.pictures['carWithBadge'].setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
                self.pictures['carWithBadge'].enabled = False
        else:
            if 'carWithBadge' in self.__carImages:
                self.pictures['carWithBadge'] = AddressableAttribute(localAddress='carWithBadge', parent=self.pictures,
                                                                     value=self.__carImages['carWithBadge'],
                                                                     valueType=Image.Image)

    def __str__(self) -> str:  # noqa: C901
        returnString: str = ''
        if self.vin.enabled and self.vin.value is not None:
            returnString += f'VIN:               {self.vin.value}\n'
        if self.model.enabled and self.model.value is not None:
            returnString += f'Model:             {self.model.value}\n'
        if self.devicePlatform.enabled and self.devicePlatform.value is not None:
            returnString += f'Device Platform:   {self.devicePlatform.value.value}\n'
        if self.nickname.enabled and self.nickname.value is not None:
            returnString += f'Nickname:          {self.nickname.value}\n'
        if self.brandCode.enabled and self.brandCode.value is not None:
            returnString += f'Brand Code:        {self.brandCode.value.value}\n'
        if self.role.enabled and self.role.value is not None:
            returnString += f'Role:              {self.role.value.value}\n'  # pylint: disable=no-member
        if self.enrollmentStatus.enabled and self.enrollmentStatus.value is not None:
            returnString += f'Enrollment Status: {self.enrollmentStatus.value.value}\n'  # pylint: disable=no-member
        if self.userRoleStatus.enabled and self.userRoleStatus.value is not None:
            returnString += f'User Role Status:  {self.userRoleStatus.value.value}\n'  # pylint: disable=no-member
        if self.coUsers.enabled:
            returnString += f'Co-Users: {len(self.coUsers)} items\n'
            for coUser in self.coUsers:
                if coUser.enabled:
                    returnString += ''.join(['\t' + line for line in str(coUser).splitlines(True)]) + '\n'
        if self.tags.enabled and self.tags.value:
            returnString += 'Tags:               ' + ', '.join(self.tags.value) + '\n'
        if self.capabilities.enabled:
            returnString += f'Capabilities: {len(self.capabilities)} items\n'
            for capability in self.capabilities.values():
                if capability.enabled:
                    returnString += ''.join(['\t' + line for line in str(capability).splitlines(True)]) + '\n'
        if self.domains.enabled:
            returnString += f'Domains: {len(self.domains)} items\n'
            for domain in self.domains:
                returnString += f'[{domain}] Elements: {len(self.domains[domain])} items\n'
                for status in self.domains[domain].values():
                    if status.enabled:
                        returnString += ''.join(['\t' + line for line in str(status).splitlines(True)]) + '\n'
                if self.domains[domain].hasError():
                    returnString += ''.join(['\t' + line for line in f'Error: {self.domains[domain].error}'.splitlines(True)]) + '\n'
        return returnString

    class Badge(Enum):
        CHARGING = 'charging'
        CONNECTED = 'connected'
        COOLING = 'cooling'
        HEATING = 'heating'
        LOCKED = 'locked'
        PARKING = 'parking'
        UNLOCKED = 'unlocked'
        VENTILATING = 'ventilating'
        WARNING = 'warning'

    class DevicePlatform(Enum,):
        MBB = 'MBB'
        MBB_ODP = 'MBB_ODP'
        MBB_OFFLINE = 'MBB_OFFLINE'
        WCAR = 'WCAR'
        UNKNOWN = 'UNKNOWN'

    class BrandCode(Enum,):
        V = 'V'
        UNKNOWN = 'unknown brand code'

    class User(AddressableObject):
        def __init__(
            self,
            localAddress: str,
            parent: AddressableObject,
            fromDict: Dict[str, str] = None,
        ) -> None:
            super().__init__(localAddress=localAddress, parent=parent)
            self.id: AddressableAttribute[str] = AddressableAttribute(localAddress='id', parent=self, value=None, valueType=str)
            self.role: AddressableAttribute[Vehicle.User.Role] = AddressableAttribute(localAddress='role', parent=self, value=None, valueType=Vehicle.User.Role)
            self.roleReseted: AddressableAttribute[bool] = AddressableAttribute(localAddress='roleReseted', parent=self, value=None, valueType=bool)
            self.enrollmentStatus: AddressableAttribute[Vehicle.User.EnrollmentStatus] = AddressableAttribute(localAddress='enrollmentStatus', parent=self,
                                                                                                              value=None,
                                                                                                              valueType=Vehicle.User.EnrollmentStatus)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict) -> None:
            LOG.debug('Update User from dict')

            if 'id' in fromDict:
                self.id.setValueWithCarTime(fromDict['id'], lastUpdateFromCar=None, fromServer=True)
            else:
                self.id.enabled = False

            if 'role' in fromDict and fromDict['role']:
                try:
                    self.role.setValueWithCarTime(Vehicle.User.Role(fromDict['role']), lastUpdateFromCar=None, fromServer=True)
                except ValueError:
                    self.role.setValueWithCarTime(Vehicle.User.Role.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('An unsupported role: %s was provided, please report this as a bug', fromDict['role'])
            else:
                self.role.enabled = False

            if 'roleReseted' in fromDict:
                self.roleReseted.setValueWithCarTime(toBool(fromDict['roleReseted']), lastUpdateFromCar=None, fromServer=True)
            else:
                self.roleReseted.enabled = False

            if 'enrollmentStatus' in fromDict and fromDict['enrollmentStatus']:
                try:
                    self.enrollmentStatus.setValueWithCarTime(Vehicle.User.EnrollmentStatus(fromDict['enrollmentStatus']), lastUpdateFromCar=None,
                                                              fromServer=True)
                except ValueError:
                    self.enrollmentStatus.setValueWithCarTime(Vehicle.User.EnrollmentStatus.UNKNOWN, lastUpdateFromCar=None, fromServer=True)
                    LOG.warning('An unsupported target operation: %s was provided, please report this as a bug', fromDict['enrollmentStatus'])
            else:
                self.enrollmentStatus.enabled = False

        def __str__(self) -> str:
            returnValue: str = ''
            if self.id.enabled and self.id.value is not None:
                returnValue += f'Id: {self.id.value}, '
            if self.role.enabled and self.role.value is not None:
                returnValue += f' Role: {self.role.value.value}, '  # pylint: disable=no-member
            if self.roleReseted.enabled and self.roleReseted.value is not None:
                returnValue += f' Reseted: {self.roleReseted.value}, '
            if self.enrollmentStatus.enabled and self.enrollmentStatus.value is not None:
                returnValue += f' Enrollment Status: {self.enrollmentStatus.value.value}'  # pylint: disable=no-member
            return returnValue

        class Role(Enum,):
            PRIMARY_USER = 'PRIMARY_USER'
            SECONDARY_USER = 'SECONDARY_USER'
            GUEST_USER = 'GUEST_USER'
            CDIS_UNKNOWN_USER = 'CDIS_UNKNOWN_USER'
            UNKNOWN = 'UNKNOWN'

            def __str__(self) -> str:
                return self.value

        class EnrollmentStatus(Enum,):
            STARTED = 'STARTED'
            NOT_STARTED = 'NOT_STARTED'
            COMPLETED = 'COMPLETED'
            GDC_MISSING = 'GDC_MISSING'
            INACTIVE = 'INACTIVE'
            UNKNOWN = 'UNKNOWN'

        class UserRoleStatus(Enum,):
            ENABLED = 'ENABLED'
            DISABLED_HMI = 'DISABLED_HMI'
            DISABLED_SPIN = 'DISABLED_SPIN'
            DISABLED_PU_SPIN_RESET = 'DISABLED_PU_SPIN_RESET'
            CDIS_UNKNOWN_USER = 'CDIS_UNKNOWN_USER'
            UNKNOWN = 'UNKNOWN'



from typing import TYPE_CHECKING, Optional, Dict, Any

import logging

from datetime import datetime

from weconnect.addressable import AddressableObject, AddressableAttribute

if TYPE_CHECKING:
    from weconnect.elements.generic_status import GenericStatus

LOG: logging.Logger = logging.getLogger("weconnect")


class Error(AddressableObject):
    def __init__(
        self,
        localAddress: str,
        parent: Optional['GenericStatus'],
        fromDict: Dict[str, Any] = None,
    ) -> None:
        super().__init__(localAddress=localAddress, parent=parent)
        self.code: AddressableAttribute[int] = AddressableAttribute(localAddress='code', parent=self, value=None, valueType=int)
        self.message: AddressableAttribute[str] = AddressableAttribute(localAddress='message', parent=self, value=None, valueType=str)
        self.group: AddressableAttribute[int] = AddressableAttribute(localAddress='group', parent=self, value=None, valueType=int)
        self.info: AddressableAttribute[str] = AddressableAttribute(localAddress='info', parent=self, value=None, valueType=str)
        self.timestamp: AddressableAttribute[str] = AddressableAttribute(localAddress='timestamp', parent=self, value=None, valueType=datetime)
        self.retry: AddressableAttribute[bool] = AddressableAttribute(localAddress='retry', parent=self, value=None, valueType=bool)

        if fromDict is not None:
            self.update(fromDict)

    def reset(self) -> None:
        if self.code.enabled:
            self.code.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.code.enabled = False
        if self.message.enabled:
            self.message.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.message.enabled = False
        if self.group.enabled:
            self.group.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.group.enabled = False
        if self.info.enabled:
            self.info.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.info.enabled = False
        if self.retry.enabled:
            self.retry.setValueWithCarTime(None, lastUpdateFromCar=None, fromServer=True)
            self.retry.enabled = False
        if self.enabled:
            self.enabled = False

    def update(self, fromDict: Dict[str, Any]) -> None:
        LOG.debug('Update Status Error from dict')

        self.code.fromDict(fromDict, 'code')
        self.message.fromDict(fromDict, 'message')
        self.group.fromDict(fromDict, 'group')
        self.info.fromDict(fromDict, 'info')
        self.timestamp.fromDict(fromDict, 'errorTimeStamp')
        self.retry.fromDict(fromDict, 'retry')

        if not self.code.enabled and not self.message.enabled and not self.code.enabled and not self.info.enabled \
                and not self.retry.enabled:
            self.enabled = False
        else:
            self.enabled = True

        for key, value in {key: value for key, value in fromDict.items()
                           if key not in ['code', 'message', 'group', 'info', 'errorTimeStamp', 'retry']}.items():
            LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

    def __str__(self) -> str:
        return f'Error {self.code.value}: {self.message.value} \n\tinfo: {self.info.value} \n\ttimestamp: {self.timestamp.value}'



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Dict, Any, List

from enum import Enum
import logging
from datetime import datetime, timedelta, timezone

if TYPE_CHECKING:
    from weconnect.elements.vehicle import Vehicle

from weconnect.util import robustTimeParse
from weconnect.addressable import AddressableObject, AddressableAttribute, AddressableList, AddressableDict
from weconnect.elements.control_operation import ControlOperation
from weconnect.elements.error import Error

LOG: logging.Logger = logging.getLogger("weconnect")


class GenericStatus(AddressableObject):
    def __init__(
        self,
        vehicle: Vehicle,
        parent: AddressableDict[str, GenericStatus],
        statusId: str,
        fromDict: Optional[Dict[str, Any]] = None,
        fixAPI: bool = True,
    ) -> None:
        self.vehicle: Vehicle = vehicle
        self.fixAPI: bool = fixAPI
        super().__init__(localAddress=statusId, parent=parent)
        self.id: str = statusId
        self.carCapturedTimestamp: AddressableAttribute[datetime] = AddressableAttribute(
            localAddress='carCapturedTimestamp', parent=self, value=None, valueType=datetime)
        self.error: Error = Error(localAddress='error', parent=self)
        self.requests: AddressableDict[GenericStatus.Request] = AddressableDict(localAddress='request', parent=self)

        if fromDict is not None:
            self.update(fromDict=fromDict)

    def hasError(self) -> bool:
        return self.error.enabled

    def hasRequests(self) -> bool:
        return len(self.requests) > 0

    def update(self, fromDict: Dict[str, Any], ignoreAttributes: Optional[List[str]] = None):  # noqa: C901
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update status from dict')

        if 'value' in fromDict:
            if 'carCapturedTimestamp' in fromDict['value']:
                carCapturedTimestamp: Optional[datetime] = robustTimeParse(fromDict['value']['carCapturedTimestamp'])
                if self.fixAPI and carCapturedTimestamp is not None:
                    # Looks like for some cars the calculation of the carCapturedTimestamp does not account for the timezone
                    # Unfortunatly it is unknown what the timezone of the car is. So the best we can do is substract 30
                    # minutes as long as the timestamp is in the future. This will create false results when the query
                    # interval is large
                    fixed: timedelta = timedelta(hours=0, minutes=0)
                    while carCapturedTimestamp > datetime.utcnow().replace(tzinfo=timezone.utc):
                        carCapturedTimestamp -= timedelta(hours=0, minutes=30)
                        fixed -= timedelta(hours=0, minutes=30)
                    if fixed > timedelta(hours=0, minutes=0):
                        LOG.warning('%s: Attribute carCapturedTimestamp was in the future. Substracted %s to fix this.'
                                    ' This is a problem of the weconnect API and might be fixed in the future',
                                    self.getGlobalAddress(), fixed)
                    if carCapturedTimestamp == datetime(year=2000, month=1, day=1, hour=0, minute=0, second=0,
                                                        tzinfo=timezone.utc):
                        carCapturedTimestamp = None

                self.carCapturedTimestamp.setValueWithCarTime(carCapturedTimestamp, lastUpdateFromCar=None, fromServer=True)
                if self.carCapturedTimestamp.value is None:
                    self.carCapturedTimestamp.enabled = False
            else:
                self.carCapturedTimestamp.setValueWithCarTime(None, fromServer=True)
                self.carCapturedTimestamp.enabled = False

            if isinstance(fromDict['value'], Dict):
                for key, value in {key: value for key, value in fromDict['value'].items()
                                   if key not in (['carCapturedTimestamp'] + ignoreAttributes)
                                   }.items():
                    LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)
        else:
            self.carCapturedTimestamp.setValueWithCarTime(None, fromServer=True)
            self.carCapturedTimestamp.enabled = False

        if 'error' in fromDict:
            self.error.update(fromDict['error'])
        else:
            self.error.reset()

        if 'requests' in fromDict:
            requestsToRemove = list(self.requests.keys())
            for request in fromDict['requests']:
                key = None
                if 'requestId' in request:
                    key = request['requestId']
                elif 'operation' in request:
                    key = request['operation']
                else:
                    key = "none"
                if key in self.requests:
                    self.requests[key].update(fromDict=request)
                    if key in requestsToRemove:
                        requestsToRemove.remove(key)
                else:
                    newRequest = GenericStatus.Request(localAddress=key, parent=self.requests, fromDict=request)
                    self.requests[key] = newRequest

            for requestKey in requestsToRemove:
                self.requests[requestKey].enabled = False
                del self.requests[requestKey]

        else:
            self.requests.clear()
            self.requests.enabled = False

        for key, value in {key: value for key, value in fromDict.items()
                           if key not in (['value', 'error', 'requests'] + ['carCapturedTimestamp'] + ignoreAttributes)}.items():
            LOG.warning('%s: Unknown element %s with value %s', self.getGlobalAddress(), key, value)

    def __str__(self) -> str:
        returnString: str = f'[{self.id}]'
        if self.carCapturedTimestamp.enabled and self.carCapturedTimestamp.value is not None:
            returnString += f' (last captured {self.carCapturedTimestamp.value.isoformat()})'
        if self.error.enabled:
            returnString += '\n\tError: ' + ''.join(['\t' + line for line in str(self.error).splitlines(True)])
        for request in self.requests.values():
            returnString += f'\n\tRequest: {request}'
        return returnString

    class Request(AddressableObject):
        def __init__(
            self,
            localAddress: str,
            parent: AddressableList[GenericStatus.Request],
            fromDict: Dict[str, Any] = None,
        ) -> None:
            super().__init__(localAddress=localAddress, parent=parent)
            self.status: AddressableAttribute[GenericStatus.Request.Status] = AddressableAttribute(localAddress='status', parent=self,
                                                                                                   value=None, valueType=GenericStatus.Request.Status)
            self.operation: AddressableAttribute[ControlOperation] = AddressableAttribute(
                localAddress='operation', parent=self, value=None, valueType=ControlOperation)
            self.body: AddressableAttribute[str] = AddressableAttribute(localAddress='body', parent=self, value=None, valueType=str)
            self.group: AddressableAttribute[int] = AddressableAttribute(localAddress='group', parent=self, value=None, valueType=int)
            self.info: AddressableAttribute[str] = AddressableAttribute(localAddress='info', parent=self, value=None, valueType=str)
            self.requestId: AddressableAttribute[str] = AddressableAttribute(localAddress='requestId', parent=self, value=None, valueType=str)
            self.vcfRequestId: AddressableAttribute[str] = AddressableAttribute(localAddress='vcfRequestId', parent=self, value=None, valueType=str)

            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict: Dict[str, Any]) -> None:  # noqa: C901
            LOG.debug('Update Request from dict')
            self.status.fromDict(fromDict, 'status')
            self.operation.fromDict(fromDict, 'operation')
            self.body.fromDict(fromDict, 'body')
            self.group.fromDict(fromDict, 'group')
            self.info.fromDict(fromDict, 'info')
            self.requestId.fromDict(fromDict, 'requestId')
            self.vcfRequestId.fromDict(fromDict, 'vcfRequestId')

        def __str__(self) -> str:
            returnValue: str = ''
            if self.operation.enabled and self.operation.value is not None:
                returnValue += f'{self.operation.value.value} operation,'
            if self.status.enabled and self.status.value is not None:
                returnValue += f' status {self.status.value.value} '
            if self.info.enabled and self.info.value is not None:
                returnValue += f' information: {self.info.value}'
            if self.requestId.enabled and self.requestId.value is not None:
                returnValue += f' Request Id: {self.requestId.value}'
            return returnValue

        class Status(Enum,):
            SUCCESSFULL = 'successful'
            FAIL = 'fail'
            POLLING_TIMEOUT = 'polling_timeout'
            IN_PROGRESS = 'in_progress'
            QUEUED = 'queued'
            DELAYED = 'delayed'
            TIMEOUT = 'timeout'
            FAIL_VEHICLE_IS_OFFLINE = 'fail_vehicle_is_offline'
            FAIL_IGNITION_ON = 'fail_ignition_on'
            FAIL_BATTERY_LOW = 'fail_battery_low'
            FAIL_PLUG_ERROR = 'fail_plug_error'
            FAIL_CHARGE_PLUG_NOT_CONNECTED = 'fail_charge_plug_not_connected'
            FAIL_NO_EXTERNAL_POWER = 'fail_no_external_power'
            UNKNOWN = 'unknown status'



from enum import Enum
import logging

from weconnect.addressable import AddressableAttribute, AddressableDict, AddressableObject
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class WindowHeatingStatus(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.windows = AddressableDict(localAddress='windows', parent=self)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update window heating status from dict')

        if 'value' in fromDict:
            if 'windowHeatingStatus' in fromDict['value'] and fromDict['value']['windowHeatingStatus'] is not None:
                for windowDict in fromDict['value']['windowHeatingStatus']:
                    if 'windowLocation' in windowDict:
                        if windowDict['windowLocation'] in self.windows:
                            self.windows[windowDict['windowLocation']].update(fromDict=windowDict)
                        else:
                            self.windows[windowDict['windowLocation']] = WindowHeatingStatus.Window(
                                fromDict=windowDict, parent=self.windows)
                for windowName in [windowName for windowName in self.windows.keys()
                                   if windowName not in [window['windowLocation']
                                   for window in fromDict['value']['windowHeatingStatus'] if 'windowLocation' in window]]:
                    del self.windows[windowName]
            else:
                self.windows.clear()
                self.windows.enabled = False
        else:
            self.windows.clear()
            self.windows.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(ignoreAttributes + ['windowHeatingStatus']))

    def __str__(self):
        string = super().__str__()
        string += f'\n\tWindows: {len(self.windows)} items'
        for window in self.windows.values():
            string += f'\n\t\t{window}'
        return string

    class Window(AddressableObject):
        def __init__(
            self,
            parent,
            fromDict=None,
        ):
            super().__init__(localAddress=None, parent=parent)
            self.windowHeatingState = AddressableAttribute(
                localAddress='windowHeatingState', parent=self, value=None,
                valueType=WindowHeatingStatus.Window.WindowHeatingState)
            if fromDict is not None:
                self.update(fromDict)

        def update(self, fromDict):
            LOG.debug('Update window from dict')

            if 'windowLocation' in fromDict:
                self.id = fromDict['windowLocation']
                self.localAddress = self.id
            else:
                LOG.error('Window is missing windowLocation attribute')

            self.windowHeatingState.fromDict(fromDict, 'windowHeatingState')

            for key, value in {key: value for key, value in fromDict.items()
                               if key not in ['windowLocation', 'windowHeatingState']}.items():
                LOG.warning('%s: Unknown attribute %s with value %s', self.getGlobalAddress(), key, value)

        def __str__(self):
            return f'{self.id}: {self.windowHeatingState.value.value}'  # pylint: disable=no-member

        class WindowHeatingState(Enum,):
            ON = 'on'
            OFF = 'off'
            INVALID = 'invalid'
            UNKNOWN = 'unknown open state'



from enum import Enum

import hashlib

import json
import logging

from requests import Session

from weconnect.auth.we_connect_session import WeConnectSession
from weconnect.auth.we_charge_session import WeChargeSession
from weconnect.auth.my_cupra_session import MyCupraSession

LOG = logging.getLogger("weconnect")


class SessionUser():
    def __init__(self, username: str, password: str) -> None:
        self.username = username
        self.password = password

    def __str__(self) -> str:
        return f'{self.username}:{self.password}'


class Service(Enum):
    WE_CONNECT = 'WeConnect'
    WE_CHARGE = 'WeCharge'
    MY_CUPRA = 'MyCupra'

    def __str__(self) -> str:
        return self.value


class SessionManager():
    def __init__(self, tokenstorefile=None) -> None:
        if tokenstorefile is not None:
            try:
                with open(tokenstorefile, 'r', encoding='utf8') as file:
                    self.tokenstore = json.load(file)
            except json.JSONDecodeError as err:
                LOG.info('Could not use token from file %s (%s)', tokenstorefile, err.msg)
                self.tokenstore = {}
            except FileNotFoundError as err:
                LOG.info('Could not use token from file %s (%s)', tokenstorefile, err)
                self.tokenstore = {}
        else:
            self.tokenstore = {}
        self.sessions = {}

    def generateHash(service: Service, sessionuser: SessionUser) -> str:
        hashstr = service.value + str(sessionuser)
        return hashlib.sha512(hashstr.encode()).hexdigest()

    def getSession(self, service: Service, sessionuser: SessionUser) -> Session:
        session = None
        if (service, sessionuser) in self.sessions:
            return self.sessions[(service, sessionuser)]

        hash: str = SessionManager.generateHash(service, sessionuser)
        if hash in self.tokenstore:
            LOG.info('Reusing tokens from previous session')
            token = self.tokenstore[hash]
        else:
            token = None

        if service == Service.WE_CONNECT:
            session = WeConnectSession(sessionuser=sessionuser, token=token)
        elif service == Service.WE_CHARGE:
            session = WeChargeSession(sessionuser=sessionuser, token=token)
        elif service == Service.MY_CUPRA:
            session = MyCupraSession(sessionuser=sessionuser, token=token)
        self.sessions[(service, sessionuser)] = session
        return session

    def saveTokenstore(self, filename: str):
        refreshedTokenstore = {}
        for sessionkey, session in self.sessions.items():
            refreshedTokenstore[SessionManager.generateHash(sessionkey[0], sessionkey[1])] = session.token
        self.tokenstore = refreshedTokenstore
        if self.tokenstore:
            try:
                with open(filename, 'w', encoding='utf8') as file:
                    json.dump(self.tokenstore, file)
                LOG.info('Writing tokenstore to file %s', filename)
            except ValueError as err:  # pragma: no cover
                LOG.info('Could not write tokenstore to file %s (%s)', filename, err)



import logging

from weconnect.addressable import AddressableAttribute
from weconnect.elements.generic_status import GenericStatus

LOG = logging.getLogger("weconnect")


class RangeMeasurements(GenericStatus):
    def __init__(
        self,
        vehicle,
        parent,
        statusId,
        fromDict=None,
        fixAPI=True,
    ):
        self.electricRange = AddressableAttribute(
            localAddress='electricRange', parent=self, value=None, valueType=int)
        self.gasolineRange = AddressableAttribute(
            localAddress='gasolineRange', value=None, parent=self, valueType=int)
        self.adBlueRange = AddressableAttribute(
            localAddress='adBlueRange', value=None, parent=self, valueType=int)
        self.dieselRange = AddressableAttribute(
            localAddress='dieselRange', value=None, parent=self, valueType=int)
        super().__init__(vehicle=vehicle, parent=parent, statusId=statusId, fromDict=fromDict, fixAPI=fixAPI)

    def update(self, fromDict, ignoreAttributes=None):
        ignoreAttributes = ignoreAttributes or []
        LOG.debug('Update battery status from dict')

        if 'value' in fromDict:
            self.electricRange.fromDict(fromDict['value'], 'electricRange')
            self.gasolineRange.fromDict(fromDict['value'], 'gasolineRange')
            self.adBlueRange.fromDict(fromDict['value'], 'adBlueRange')
            self.dieselRange.fromDict(fromDict['value'], 'dieselRange')
        else:
            self.electricRange.enabled = False
            self.gasolineRange.enabled = False
            self.adBlueRange.enabled = False
            self.dieselRange.enabled = False

        super().update(fromDict=fromDict, ignoreAttributes=(
            ignoreAttributes + ['electricRange', 'gasolineRange', 'adBlueRange', 'dieselRange']))

    def __str__(self):
        string = super().__str__()
        if self.electricRange.enabled:
            string += f'\n\tElectric Range: {self.electricRange.value}km'
        if self.gasolineRange.enabled:
            string += f'\n\tGasoline Range: {self.gasolineRange.value}km'
        if self.adBlueRange.enabled:
            string += f'\n\tAdBlue Range: {self.adBlueRange.value}km'
        if self.dieselRange.enabled:
            string += f'\n\tDiesel Range: {self.dieselRange.value}km'
        return string



from __future__ import annotations
from enum import Enum
from typing import Any

import re
from datetime import datetime

import json

import logging

import shutil

SUPPORT_IMAGES = False
try:
    from PIL import Image  # type: ignore
    SUPPORT_IMAGES = True
except ImportError:
    pass

SUPPORT_ASCII_IMAGES = False
try:
    import ascii_magic  # type: ignore
    SUPPORT_ASCII_IMAGES = True
except ImportError:
    pass


def robustTimeParse(timeString: str) -> datetime:
    timeString = timeString.replace('Z', '+00:00')
    match = re.search(
        r'^(?P<start>\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.)(?P<fractions>\d+)(?P<end>\+\d{2}:\d{2})$', timeString)
    if match:
        timeString = match.group('start') + match.group('fractions').ljust(6, "0") + match.group('end')
    return datetime.fromisoformat(timeString).replace(microsecond=0)


def toBool(value: Any) -> bool:
    if value in [True, 'True', 'true', 'yes']:
        return True
    if value in [False, 'False', 'false', 'no']:
        return False
    raise ValueError('Not a valid boolean value (True/False)')


if SUPPORT_ASCII_IMAGES:
    def imgToASCIIArt(img: Image, columns: int = 0, mode: ascii_magic.Modes = ascii_magic.Modes.TERMINAL) -> str:
        bbox = img.getbbox()

        # Crop the image to the contents of the bounding box
        image = img.crop(bbox)

        # Determine the width and height of the cropped image
        (width, height) = image.size

        # Create a new image object for the output image
        cropped_image = Image.new("RGBA", (width, height), (0, 0, 0, 0))

        # Paste the cropped image onto the new image
        cropped_image.paste(image, (0, 0))

        if columns == 0:
            columns = shutil.get_terminal_size()[0]

        return ascii_magic.from_image(cropped_image, columns=columns, mode=mode)


def celsiusToKelvin(value):
    return value + 273.15


def kelvinToCelsius(value):
    return value - 273.15


def farenheitToKelvin(value):
    return 273.5 + ((value - 32.0) * (5.0 / 9.0))


class DuplicateFilter(logging.Filter):

    def __init__(self, name: str = '') -> None:
        super().__init__(name=name)
        self.lastLog = {}
        self.firstTime = True

    def filter(self, record) -> bool:
        # don't filter critical or error messages:
        if record.levelno >= logging.ERROR:
            return True

        if record.module in self.lastLog:
            if record.levelno in self.lastLog[record.module]:
                if self.lastLog[record.module][record.levelno] == (record.msg, record.args):
                    if self.firstTime:
                        self.firstTime = False
                        logging.info('Repeated log messages from the same module are hidden (does not apply to errors or critical problems)')
                    return False
            self.lastLog[record.module][record.levelno] = (record.msg, record.args)
        else:
            self.lastLog[record.module] = {record.levelno: (record.msg, record.args)}
        return True


class ExtendedEncoder(json.JSONEncoder):
    """Datetime object encode used for json serialization"""
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def default(self, o: Any) -> str:
        """Serialize datetime object to isodate string

        Args:
            o (datetime): datetime object

        Returns:
            str: object represented as isoformat string
        """
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, Enum):
            return o.value
        return super().default(o)


class ExtendedWithNullEncoder(ExtendedEncoder):
    """Datetime object encode used for json serialization"""

    def default(self, o: Any) -> str:
        try:
            return super().default(o)
        except TypeError:
            return None



from enum import Enum


class ControlInputEnum(Enum):
    @classmethod
    def allowedValues(cls):
        return []


class ControlOperation(ControlInputEnum):
    START = 'start'
    STOP = 'stop'
    NONE = 'none'
    SETTINGS = 'settings'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [ControlOperation.START, ControlOperation.STOP]


class AccessControlOperation(ControlInputEnum):
    LOCK = 'lock'
    UNLOCK = 'unlock'
    NONE = 'none'
    UNKNOWN = 'unknown'

    @classmethod
    def allowedValues(cls):
        return [AccessControlOperation.LOCK, AccessControlOperation.UNLOCK]
